import copy
import json
import logging
import os

from command.model.configuration import *
from utils import exceptions
from utils.base64 import b64encode_str
from utils.case import to_camel_case
from .cfg_reader import CfgReader
from .workspace_helper import ArgumentUpdateMixin

logger = logging.getLogger('backend')


class WorkspaceCfgEditor(CfgReader, ArgumentUpdateMixin):

    @staticmethod
    def get_cfg_folder(ws_folder, resource_id):
        path = os.path.join(ws_folder, "Resources")
        name = b64encode_str(resource_id)
        while len(name):
            if len(name) > 255:
                path = os.path.join(path, name[:254]+'+')
                name = name[254:]
            else:
                path = os.path.join(path, name)
                name = ""
        return path

    @classmethod
    def get_cfg_path(cls, ws_folder, resource_id):
        return os.path.join(cls.get_cfg_folder(ws_folder, resource_id), f"cfg.json")

    @classmethod
    def load_resource(cls, ws_folder, resource_id, version):
        path = cls.get_cfg_path(ws_folder, resource_id)
        with open(path, 'r', encoding="utf-8") as f:
            data = json.load(f)
        if '$ref' in data:
            ref_resource_id = data['$ref']
            path = cls.get_cfg_path(ws_folder, ref_resource_id)
            with open(path, 'r', encoding="utf-8") as f:
                data = json.load(f)
        cfg = CMDConfiguration(data)
        for resource in cfg.resources:
            if resource.version != version:
                raise ValueError(f"Resource version not match: {version} != {resource.version}")
        cfg_editor = cls(cfg)
        cfg_editor.reformat()
        return cfg_editor

    @classmethod
    def new_cfg(cls, plane, resources, command_groups):
        assert len(resources) and len(command_groups)
        cfg = CMDConfiguration()
        cfg.plane = plane
        cfg.resources = resources
        cfg.command_groups = command_groups
        cfg_editor = cls(cfg)
        cfg_editor.reformat()
        return cfg_editor

    def __init__(self, cfg, deleted=False):
        super().__init__(cfg)
        self.deleted = deleted

    def iter_cfg_files_data(self):
        if self.deleted:
            for resource in self.resources:
                yield resource.id, None
        else:
            for resource_id, data in super().iter_cfg_files_data():
                yield resource_id, data

    def rename_command_group(self, *cg_names, new_cg_names):
        if len(cg_names) < 1:
            raise exceptions.InvalidAPIUsage(f"Invalid command group name, it's empty")
        if len(new_cg_names) < 1:
            raise exceptions.InvalidAPIUsage(f"Invalid new command group name, it's empty")

        cg_names = [*cg_names]
        new_cg_names = [*new_cg_names]

        group, tail_names, parent, remain_names = self.find_command_group(*cg_names)

        if not group:
            raise exceptions.InvalidAPIUsage(f"Cannot find command group name '{' '.join(cg_names)}'")

        # remove command group from parent command group
        idx = parent.command_groups.index(group)
        parent.command_groups.pop(idx)

        new_group_names = [*new_cg_names, *tail_names]

        conflict_group, conflict_tail_names, parent, remain_names = self.find_command_group(*new_group_names)
        if not conflict_group:
            if parent.command_groups is None:
                parent.command_groups = []
            group.name = ' '.join(remain_names)
            parent.command_groups.append(group)
        elif conflict_tail_names:
            idx = parent.command_groups.index(conflict_group)
            parent.command_groups.pop(idx)
            group.name = ' '.join(remain_names)
            conflict_group.name = ' '.join(conflict_tail_names)
            if group.command_groups is None:
                group.command_groups = []
            group.command_groups.append(conflict_group)
            parent.command_groups.append(group)
        else:
            assert group.commands or group.command_groups
            raise exceptions.InvalidAPIUsage(f"Command Group '{' '.join(new_group_names)}' already exist")

        self.reformat()

    def rename_command(self, *cmd_names, new_cmd_names):
        command = self._remove_command(*cmd_names)
        command.name = new_cmd_names[-1]
        self._add_command(*new_cmd_names, command=command)
        self.reformat()

    def _remove_command(self, *cmd_names):
        if len(cmd_names) < 2:
            raise exceptions.InvalidAPIUsage(f"Invalid command name, it's empty")

        cmd_names = [*cmd_names]
        command = self.find_command(*cmd_names)
        if command is None:
            raise exceptions.ResourceNotFind(f"Cannot find definition for command '{' '.join(cmd_names)}'")

        command_group, tail_names, _, _ = self.find_command_group(*cmd_names[:-1])
        assert command_group is not None and not tail_names

        idx = command_group.commands.index(command)
        command_group.commands.pop(idx)

        return command

    def _add_command(self, *cmd_names, command):
        if len(cmd_names) < 2:
            raise exceptions.InvalidAPIUsage(f"Invalid new command name, it's empty")
        cmd_names = [*cmd_names]

        if self.find_command(*cmd_names):
            raise exceptions.InvalidAPIUsage(f"Command '{' '.join(cmd_names)}' already exist")

        command_group, tail_names, parent, remain_names = self.find_command_group(*cmd_names[:-1])
        if not command_group:
            cg_names = remain_names
            while not command_group and len(cg_names) > 1:
                cg_names = cg_names[:-1]
                command_group, tail_names, _, _ = self.find_command_group(*cg_names, parent=parent)
            if command_group:
                assert tail_names
                idx = parent.command_groups.index(command_group)
                parent.command_groups.pop(idx)
                new_command_group = CMDCommandGroup()
                new_command_group.name = ' '.join(cg_names)
                command_group.name = ' '.join(tail_names)
                new_command_group.command_groups = [command_group]
                parent.command_groups.append(new_command_group)
                parent = new_command_group
                remain_names = remain_names[len(cg_names):]

            new_command_group = CMDCommandGroup()
            new_command_group.name = ' '.join(remain_names)
            if parent.command_groups is None:
                parent.command_groups = []
            parent.command_groups.append(new_command_group)
            command_group = new_command_group

        elif tail_names:
            idx = parent.command_groups.index(command_group)
            parent.command_groups.pop(idx)
            new_command_group = CMDCommandGroup()
            new_command_group.name = ' '.join(remain_names)
            command_group.name = ' '.join(tail_names)
            new_command_group.command_groups = [command_group]
            parent.command_groups.append(new_command_group)
            command_group = new_command_group

        if command_group.commands is None:
            command_group.commands = []

        command_group.commands.append(command)

    def merge(self, plus_cfg_editor):
        if not self._can_merge(plus_cfg_editor):
            return None

        _, plus_command = [*plus_cfg_editor.iter_commands_by_operations('get')][0]
        plus_op_required_args, plus_op_optional_args = plus_cfg_editor._parse_command_http_op_url_args(plus_command)

        main_editor = WorkspaceCfgEditor(self.cfg.__class__(self.cfg.to_primitive()))  # generate a copy of main cfg
        main_commands = [command for _, command in main_editor.iter_commands_by_operations('get')]
        for main_command in main_commands:
            # merge args
            new_args = set()
            for args in plus_op_required_args.values():
                new_args.update(args)
            for args in plus_op_optional_args.values():
                new_args.update(args)

            for arg_group in plus_command.arg_groups:
                arg_group = plus_cfg_editor._filter_args_in_arg_group(arg_group, new_args, copy=True)
                if arg_group:
                    try:
                        main_editor._command_merge_arg_group(main_command, arg_group)
                    except exceptions.InvalidAPIUsage as ex:
                        logger.error(ex)
                        return None

            # create conditions
            main_op_required_args, _ = main_editor._parse_command_http_op_url_args(main_command)
            plus_operations = []
            for operation in plus_command.operations:
                plus_operations.append(operation.__class__(operation.to_primitive()))
            op_required_args = {**plus_op_required_args, **main_op_required_args}
            common_required_args, main_command.conditions, main_command.operations = main_editor._merge_command_operations(
                op_required_args,
                *plus_operations, *main_command.operations
            )

            # update arg required of command
            for arg_group in main_command.arg_groups:
                for arg in arg_group.args:
                    arg.required = arg.var in common_required_args

            for resource in plus_command.resources:
                main_command.resources.append(
                    resource.__class__(resource.to_primitive())
                )

            # relink main_command
            main_command.link()

        for resource in plus_cfg_editor.resources:
            main_editor.cfg.resources.append(
                resource.__class__(resource.to_primitive())
            )

        main_editor.reformat()
        return main_editor

    def update_command_confirmation(self, *cmd_names, confirmation):
        if len(cmd_names) < 2:
            raise exceptions.InvalidAPIUsage(f"Invalid command name, it's empty")

        command = self.find_command(*cmd_names)
        if command is None:
            raise exceptions.ResourceNotFind(f"Cannot find definition for command '{' '.join(cmd_names)}'")
        if confirmation != command.confirmation:
            command.confirmation = confirmation
        self.reformat()
    
    def update_command_outputs(self, *cmd_names, outputs):
        if len(cmd_names) < 2:
            raise exceptions.InvalidAPIUsage(f"Invalid command name, it's empty")

        command = self.find_command(*cmd_names)
        if command is None:
            raise exceptions.ResourceNotFind(f"Cannot find definition for command '{' '.join(cmd_names)}'")
        if not outputs:
            command.outputs = None
        else:
            command.outputs = CMDCommand(raw_data={"outputs": outputs}).outputs
            try:
                for output in command.outputs:
                    output.validate()
            except Exception as err:
                raise exceptions.InvalidAPIUsage(f"Invalid output data: {err}")
        self.reformat()

    def update_arg_by_var(self, *cmd_names, arg_var, **kwargs):
        arg, _ = self.find_arg_by_var(*cmd_names, arg_var=arg_var)
        if not arg:
            return None
        self._update_arg(arg, **kwargs)
        self.reformat()

    def unwrap_cls_arg(self, *cmd_names, arg_var):
        command = self.find_command(*cmd_names)
        self._unwrap_cls_arg_in_command(command, arg_var)
        self.reformat()

    @classmethod
    def _unwrap_cls_arg_in_command(cls, command, arg_var):
        parent_arg, arg, _ = cls.find_arg_in_command_with_parent_by_var(command, arg_var=arg_var)
        if not arg:
            raise exceptions.InvalidAPIUsage(
                f"Argument not exist: {arg.var}")
        assert parent_arg is not None

        if isinstance(arg, CMDObjectArgBase):
            if not arg.cls:
                raise exceptions.InvalidAPIUsage(
                    f"{arg.var} is not a class argument"
                )
        elif not isinstance(arg, CMDClsArgBase):
            raise exceptions.InvalidAPIUsage(
                f"{arg.var} is not a class argument"
            )

        linked_schema_parent = None
        linked_schema = None

        # find linked schema
        for parent_schema, schema, _ in cls.iter_schema_in_command_by_arg_var(command, arg_var=arg_var):
            if linked_schema is not None:
                raise exceptions.InvalidAPIUsage(
                    f"Cannot unwrap argument: {arg.var} is used by mutiple schemas."
                )
            linked_schema_parent = parent_schema
            linked_schema = schema

        if linked_schema is None:
            raise exceptions.InvalidAPIUsage(
                f"Cannot unwrap argument: {arg.var} is not used by any schema."
            )

        if isinstance(linked_schema, CMDClsSchemaBase):
            new_schema = linked_schema.get_unwrapped()
            assert new_schema.cls == linked_schema.type[1:]
            new_schema.cls = None  # set the cls to None
            cls.replace_schema(linked_schema_parent, linked_schema, new_schema)
        elif isinstance(linked_schema, (CMDObjectSchemaBase, CMDArraySchemaBase)):
            assert linked_schema.cls is not None
            # unwrap one clsSchema to instance
            for ref_parent, ref_schema, _ in cls.iter_schema_cls_reference(command, linked_schema.cls):
                assert isinstance(ref_schema, CMDClsSchemaBase)
                new_ref_schema = ref_schema.get_unwrapped()
                assert new_ref_schema.cls == linked_schema.cls
                cls.replace_schema(ref_parent, ref_schema, new_ref_schema)
                break
            # unregister current linked schema cls
            linked_schema.cls = None
        else:
            raise exceptions.InvalidAPIUsage(
                f"Not supported schema type: {type(linked_schema)}"
            )

        # regenerate args and its relationship with schema
        command._reformat_operations()  # handle duplicated schema cls definition
        command.generate_args()
        command.link()

    @staticmethod
    def replace_schema(parent, schema, new_schema):
        if isinstance(parent, (CMDObjectSchemaBase, CMDObjectSchemaDiscriminator)):
            if isinstance(schema, CMDSchema):
                assert isinstance(new_schema, CMDSchema)
                schema_idx = None
                for idx in range(len(parent.props)):
                    if parent.props[idx].name == schema.name:
                        schema_idx = idx
                        break
                assert schema_idx is not None
                parent.props[schema_idx] = new_schema
            else:
                assert isinstance(parent, CMDObjectSchemaBase) and isinstance(schema, CMDSchemaBase)
                assert not isinstance(new_schema, CMDSchema) and isinstance(new_schema, CMDSchemaBase)
                assert parent.additional_props.item == schema
                parent.additional_props.item = new_schema
        elif isinstance(parent, CMDArraySchemaBase):
            assert isinstance(schema, CMDSchemaBase)
            assert not isinstance(new_schema, CMDSchema) and isinstance(new_schema, CMDSchemaBase)
            assert parent.item == schema
            parent.item = new_schema
        else:
            raise NotImplementedError()

    def flatten_arg(self, *cmd_names, arg_var, sub_args_options=None):
        command = self.find_command(*cmd_names)
        parent, arg, _ = self.find_arg_with_parent_by_var(*cmd_names, arg_var=arg_var)
        if not arg:
            raise exceptions.InvalidAPIUsage(
                f"Argument not exist: {arg.var}")
        assert parent is not None

        need_unwrap = False
        if isinstance(arg, CMDClsArg) and isinstance(arg.implement, CMDObjectArg):
            if arg.implement.additional_props:
                raise exceptions.InvalidAPIUsage(f"Cannot flatten argument with additional properties")
            need_unwrap = True
        elif isinstance(arg, CMDObjectArg):
            if arg.additional_props:
                raise exceptions.InvalidAPIUsage(f"Cannot flatten argument with additional properties")
            if arg.cls:
                need_unwrap = True
        else:
            raise exceptions.InvalidAPIUsage(f"Cannot flatten argument in type: '{type(arg)}'")

        if need_unwrap:
            # unwrap cls argument first
            self.unwrap_cls_arg(*cmd_names, arg_var=arg_var)
            command = self.find_command(*cmd_names)
            parent, arg, _ = self.find_arg_with_parent_by_var(*cmd_names, arg_var=arg_var)
            assert arg and parent

        parent.args.remove(arg)

        used_options = set()
        if isinstance(parent, CMDArgGroup):
            for group in command.arg_groups:
                for a in group.args:
                    used_options.update(a.options)
        else:
            for a in parent.args:
                used_options.update(a.options)

        for sub_arg in arg.args:
            if sub_args_options and sub_arg.var in sub_args_options:
                sub_arg.options = sub_args_options[sub_arg.var]
            # verify duplicated option
            for option in sub_arg.options:
                if option in used_options:
                    raise exceptions.VerificationError(
                        message=f"Argument option '{option}' duplicated.",
                        details={
                            "type": "Argument",
                            "options": sub_arg.options,
                            "var": sub_arg.var,
                        }
                    )
                used_options.add(option)

            sub_arg.group = to_camel_case(arg.options[0])
            sub_arg.hide = sub_arg.hide or arg.hide  # inherient the hide property from the parent arguments
            parent.args.append(sub_arg)

        # regenerate args and its relationship with schema
        command.generate_args()
        self.reformat()

    def unflatten_arg(self, *cmd_names, arg_var, options, help, sub_args_options=None):
        command = self.find_command(*cmd_names)
        parent, arg, _ = self.find_arg_with_parent_by_var(*cmd_names, arg_var=arg_var)
        if arg:
            raise exceptions.InvalidAPIUsage(
                f"Argument already exist: {arg_var}")
        if not parent:
            assert not arg
            raise exceptions.InvalidAPIUsage(
                f"Sub arguments not exist: {arg_var}")

        sub_args = []
        args = []
        for a in parent.args:
            if a.var.startswith(f'{arg_var}.'):
                a.group = None
                if sub_args_options and a.var in sub_args_options:
                    a.options = sub_args_options[a.var]
                sub_args.append(a)
            else:
                args.append(a)
        assert sub_args

        new_arg = CMDObjectArg()
        new_arg.var = arg_var
        new_arg.options = options
        new_arg.help = CMDArgumentHelp(raw_data=help)
        new_arg.args = sub_args

        args.append(new_arg)
        parent.args = args

        # regenerate args and its relation ship with schema
        command.generate_args()
        self.reformat()

    def reformat(self):
        self.cfg.reformat()

    def _parse_command_http_op_url_args(self, command):
        operation_required_args = {}
        operation_optional_args = {}
        for http_op in command.operations:
            if not isinstance(http_op, CMDHttpOperation):
                continue
            required_args, optional_args = self.parse_http_operation_url_args(http_op)
            operation_required_args[http_op.operation_id] = required_args
            operation_optional_args[http_op.operation_id] = optional_args
        return operation_required_args, operation_optional_args

    @staticmethod
    def parse_http_operation_url_args(op):
        required_args = set()
        optional_args = set()
        request = op.http.request
        assert isinstance(request, CMDHttpRequest)
        if request.path and request.path.params:
            for param in request.path.params:
                if param.arg:
                    if param.required:
                        required_args.add(param.arg)
                    else:
                        optional_args.add(param.arg)
        if request.query and request.query.params:
            for param in request.query.params:
                if param.arg:
                    if param.required:
                        required_args.add(param.arg)
                    else:
                        optional_args.add(param.arg)
        return required_args, optional_args

    def _can_merge(self, plus_cfg_editor):
        if len([*plus_cfg_editor.iter_commands()]) != 1:
            return False
        plus_commands = [command for _, command in plus_cfg_editor.iter_commands_by_operations('get')]
        if len(plus_commands) != 1:
            return False
        plus_command = plus_commands[0]
        if len(plus_command.resources) != len(plus_cfg_editor.resources):
            return False

        main_get_commands = [command for _, command in self.iter_commands_by_operations('get')]
        if len(main_get_commands) == 0:
            return False
        for command in main_get_commands:
            if len(command.resources) != len(self.resources):
                return False

        plus_200_response = None
        for http_op in plus_command.operations:
            if not isinstance(http_op, CMDHttpOperation):
                continue
            assert http_op.http.request.method == 'get'
            for response in http_op.http.responses:
                if response.is_error:
                    continue
                if 200 in response.status_codes:
                    plus_200_response = response
                    break
            if plus_200_response:
                break
        if not plus_200_response:
            return False

        main_200_response = None
        for command in main_get_commands:
            for http_op in command.operations:
                if not isinstance(http_op, CMDHttpOperation):
                    continue
                assert http_op.http.request.method == 'get'
                for response in http_op.http.responses:
                    if response.is_error:
                        continue
                    if 200 in response.status_codes:
                        if plus_200_response.diff(response, CMDDiffLevelEnum.Structure):
                            return False
                        main_200_response = response
        if not main_200_response:
            return False

        plus_op_required_args, plus_op_optional_args = self._parse_command_http_op_url_args(plus_command)
        for main_command in main_get_commands:
            main_op_required_args, main_op_optional_args = self._parse_command_http_op_url_args(main_command)
            for main_op_id, main_required_args in main_op_required_args.items():
                if main_op_id in plus_op_required_args:
                    # the operation id should be different with plus's
                    return False
                for plus_required_args in plus_op_required_args.values():
                    if plus_required_args == main_required_args:
                        # the required arguments should be different
                        return False
        return True

    def _filter_args_in_arg_group(self, arg_group, arg_vars, copy=True):
        assert isinstance(arg_group, CMDArgGroup)
        if copy:
            arg_group = arg_group.__class__(arg_group.to_primitive())
        args = []
        for arg in arg_group.args:
            if arg.var in arg_vars:
                args.append(arg)
            elif isinstance(arg, CMDObjectArg):
                obj_arg = self._filter_args_in_object_arg(arg, arg_vars, copy=False)
                if obj_arg:
                    args.append(obj_arg)
            elif isinstance(arg, CMDArrayArg):
                arr_arg = self._filter_args_in_array_arg(arg, arg_vars, copy=False)
                if arr_arg:
                    args.append(arr_arg)
        if args:
            arg_group.args = args
            return arg_group
        return None

    def _filter_args_in_array_arg(self, array_arg, arg_vars, copy=True):
        assert isinstance(array_arg, CMDArrayArgBase)
        if copy:
            array_arg = array_arg.__class__(array_arg.to_primitive())
        item = self._filter_args_in_item(array_arg.item, arg_vars, copy=False)
        if item:
            array_arg.item = item
            return array_arg
        return None

    def _filter_args_in_item(self, item, arg_vars, copy=True):
        if isinstance(item, CMDObjectArgBase):
            return self._filter_args_in_object_arg(item, arg_vars, copy)
        elif isinstance(item, CMDArrayArgBase):
            return self._filter_args_in_array_arg(item, arg_vars, copy)
        return None

    def _filter_args_in_object_arg(self, object_arg, arg_vars, copy=True):
        assert isinstance(object_arg, CMDObjectArgBase)
        if copy:
            object_arg = object_arg.__class__(object_arg.to_primitive())
        contains = False
        if object_arg.args:
            args = []
            for arg in object_arg.args:
                if arg.var in arg_vars:
                    args.append(arg)
                elif isinstance(arg, CMDObjectArg):
                    obj_arg = self._filter_args_in_object_arg(arg, arg_vars, copy=False)
                    if obj_arg:
                        args.append(obj_arg)
                elif isinstance(arg, CMDArrayArg):
                    arr_arg = self._filter_args_in_array_arg(arg, arg_vars, copy=False)
                    if arr_arg:
                        args.append(arr_arg)
            if args:
                object_arg.args = args
                contains = True

        if object_arg.additional_props:
            additional_props = None
            if object_arg.additional_props.item:
                item = self._filter_args_in_item(object_arg.additional_props.item, arg_vars, copy=False)
                if item:
                    additional_props = object_arg.additional_props
                    additional_props.item = item
            if additional_props:
                object_arg.additional_props = additional_props
                contains = True

        if contains:
            return object_arg
        return None

    def _command_merge_arg_group(self, command, arg_group):
        assert isinstance(arg_group, CMDArgGroup)
        remain_args = []
        for arg in arg_group.args:
            for cmd_ag in command.arg_groups:
                arg = self._arg_group_merge_arg(cmd_ag, arg)
                if not arg:
                    break
            if arg:
                remain_args.append(arg)
        if remain_args:
            arg_group.args = remain_args
            match_ag = None
            for ag in command.arg_groups:
                if ag.name == arg_group.name:
                    match_ag = ag
                    break
            if match_ag:
                match_ag.args.extend(arg_group.args)
            else:
                command.arg_groups.append(arg_group)

    def _arg_group_merge_arg(self, arg_group, arg):
        for cmd_arg in arg_group.args:
            if cmd_arg.var == arg.var and not isinstance(arg, type(cmd_arg)):
                raise exceptions.InvalidAPIUsage(
                    f"Same Arg var but different arg types: {arg.var} : {type(cmd_arg)} != {type(arg)}")
            if cmd_arg.var == arg.var:
                return None
            # TODO: handle merge complex args
        return arg

    def _merge_command_operations(self, op_required_args, *operations):
        common_required_args = None
        for required_args in op_required_args.values():
            if common_required_args is None:
                common_required_args = {*required_args}
            else:
                common_required_args.intersection_update(required_args)

        arg_ops_map = {}
        for op_id, required_args in op_required_args.items():
            for arg in required_args:
                if arg not in arg_ops_map:
                    arg_ops_map[arg] = set()
                arg_ops_map[arg].add(op_id)
        conditions = []
        new_operations = []
        for operation in operations:
            assert operation.operation_id in op_required_args
            op_id = operation.operation_id

            has_args = [*op_required_args[operation.operation_id]]
            if has_args:
                conflict_ops_set = set(arg_ops_map[has_args[0]])
                for arg in has_args[1:]:
                    conflict_ops_set.intersection_update(arg_ops_map[arg])
                assert op_id in conflict_ops_set
                conflict_ops_set.remove(op_id)
                not_has_args = set()
                for c_op_id in conflict_ops_set:
                    not_has_args.update(op_required_args[c_op_id])
                not_has_args.difference_update(has_args)
            else:
                not_has_args = set(arg_ops_map.keys())

            # generate condition
            assert len(has_args) + len(not_has_args) > 0
            condition = CMDCondition()
            condition.var = f"$Condition_{op_id}"
            condition.operator = CMDConditionAndOperator()
            condition.operator.operators = []
            for has_arg in sorted(has_args):
                has_operator = CMDConditionHasValueOperator()
                has_operator.arg = has_arg
                condition.operator.operators.append(has_operator)
            for not_has_arg in sorted(not_has_args):
                not_operator = CMDConditionNotOperator()
                not_operator.operator = CMDConditionHasValueOperator()
                not_operator.operator.arg = not_has_arg
                condition.operator.operators.append(not_operator)
            conditions.append(condition)

            operation.when = [condition.var]
            new_operations.append(operation)

        return common_required_args, conditions, new_operations

    def inherit_modification(self, ref_cfg: CfgReader):
        command_rename_list = []
        existing_resources = set()
        existing_sub_resources = set()
        for cmd_names, command in self.iter_commands():
            counterpart = self._find_ref_command_counterpart(cmd_names, command, ref_cfg)
            if not counterpart:
                continue
            ref_cmd_names, ref_command = counterpart
            self._inherit_modification_in_command(command, ref_command)
            command_rename_list.append((cmd_names, ref_cmd_names))
            for r in ref_command.resources:
                existing_resources.add(r.id)
                if r.subresource:
                    existing_sub_resources.add((r.id, r.subresource))

        # rename commands
        for cmd_names, ref_cmd_names in command_rename_list:
            self.rename_command(*cmd_names, new_cmd_names=ref_cmd_names)

        # inherit sub command
        sub_resources = set()
        array_sub_resources = set()
        dict_sub_resources = set()
        for ref_cmd_names, ref_command in ref_cfg.iter_commands():
            for r in ref_command.resources:
                if r.id not in existing_resources:
                    # ignore unrelated resources
                    continue
                if not r.subresource:
                    continue
                subresource_id = r.subresource
                if subresource_id.endswith('[]'):
                    subresource_id = subresource_id[:-2]
                    array_sub_resources.add((r.id, subresource_id))
                elif subresource_id.endswith('{}'):
                    subresource_id = subresource_id[:-2]
                    dict_sub_resources.add((r.id, subresource_id))
                sub_resources.add((r.id, subresource_id))
        sub_resources.difference_update(existing_sub_resources)
        if not sub_resources:
            return

        command_rename_list = []
        for resource_id, subresource_id in sub_resources:
            update_cmd_info = self.get_update_cmd(resource_id)
            if not update_cmd_info:
                continue
            update_cmd_names, update_cmd, update_by = update_cmd_info
            if update_by != "GenericOnly":
                continue

            update_cmd.link()
            update_op = None
            for operation in update_cmd.operations:
                if isinstance(operation, CMDInstanceUpdateOperation):
                    update_op = operation
            assert update_op is not None
            update_json = update_op.instance_update.json
            subresource_idx = self.idx_to_list(subresource_id)
            schema = self.find_schema_in_json(update_json, subresource_idx)
            if not schema:
                # schema not exist
                continue
            assert isinstance(schema, CMDSchema)

            # build ref_args_options
            cg_names = None
            ref_args_options = {}
            for ref_cmd_names, ref_command in ref_cfg.iter_commands_by_resource(resource_id, subresource_id):
                cg_names = ref_cmd_names[:-1]
                for ref_arg_group in ref_command.arg_groups:
                    for ref_arg in ref_arg_group.args:
                        ref_args_options[ref_arg.var] = [*ref_arg.options]
            if (resource_id, subresource_id) in array_sub_resources:
                for ref_cmd_names, ref_command in ref_cfg.iter_commands_by_resource(resource_id, subresource_id + '[]'):
                    cg_names = ref_cmd_names[:-1]
                    for ref_arg_group in ref_command.arg_groups:
                        for ref_arg in ref_arg_group.args:
                            ref_args_options[ref_arg.var] = [*ref_arg.options]
            if (resource_id, subresource_id) in dict_sub_resources:
                for ref_cmd_names, ref_command in ref_cfg.iter_commands_by_resource(resource_id, subresource_id + '{}'):
                    cg_names = ref_cmd_names[:-1]
                    for ref_arg_group in ref_command.arg_groups:
                        for ref_arg in ref_arg_group.args:
                            ref_args_options[ref_arg.var] = [*ref_arg.options]
            assert cg_names is not None

            # generate sub commands
            sub_commands = self._generate_sub_commands(schema, subresource_idx, update_cmd, ref_args_options)
            for sub_command in sub_commands:
                cmd_names = [*cg_names, sub_command.name]
                self._add_command(*cmd_names, command=sub_command)
                counterpart = self._find_ref_command_counterpart(cmd_names, sub_command, ref_cfg)
                if not counterpart:
                    continue
                ref_cmd_names, ref_command = counterpart
                self._inherit_modification_in_command(sub_command, ref_command)
                command_rename_list.append((cmd_names, ref_cmd_names))

        # rename sub commands
        for cmd_names, ref_cmd_names in command_rename_list:
            self.rename_command(*cmd_names, new_cmd_names=ref_cmd_names)

        self.reformat()

    @staticmethod
    def _find_ref_command_counterpart(cmd_names, command, ref_cfg):
        counterpart = None

        # find counterpart
        ops_methods = set()
        for operation in command.operations:
            if isinstance(operation, CMDInstanceUpdateOperation):
                ops_methods.add('instance-update')
            elif isinstance(operation, CMDInstanceCreateOperation):
                ops_methods.add('instance-create')
            elif isinstance(operation, CMDInstanceDeleteOperation):
                ops_methods.add('instance-delete')
            elif isinstance(operation, CMDHttpOperation):
                ops_methods.add(operation.http.request.method.lower())
        for ref_cmd_names, ref_command in ref_cfg.iter_commands_by_operations(*ops_methods):
            command_resources = {(r.id, r.subresource) for r in command.resources}
            ref_command_resources = {(r.id, r.subresource) for r in ref_command.resources}
            if not command_resources.issubset(ref_command_resources):
                # resources not match
                continue
            if counterpart:
                raise exceptions.ResourceConflict(
                    message=f"Failed to inherit modification for command: '{' '.join(cmd_names)}', multiple reference commands find: '{' '.join(counterpart[0])}' & '{' '.join(ref_cmd_names)}'"
                )
            counterpart = (ref_cmd_names, ref_command)
        return counterpart

    @classmethod
    def _inherit_modification_in_command(cls, command, ref_command):
        # confirmation
        if ref_command.confirmation is not None:
            command.confirmation = ref_command.confirmation

        # unwrap modification
        arg_cls_names = set()
        for _, arg, _, arg_var in cls._iter_arg_cls_definition(command):
            arg_cls_names.add(arg.cls)

        for cls_name in arg_cls_names:
            unwrap_detect = True
            while unwrap_detect:
                _, _, _, arg_var = cls._find_arg_cls_definition(command, cls_name=cls_name)
                if not arg_var:
                    break

                arg_vars = [arg_var]
                for _, _, _, arg_var in cls._iter_arg_cls_reference(command, cls_name=cls_name):
                    arg_vars.append(arg_var)

                unwrap_arg_vars = set()
                for arg_var in arg_vars:
                    schema_idx = None
                    for _, _, s_idx in cls.iter_schema_in_command_by_arg_var(command, arg_var):
                        assert schema_idx is None
                        schema_idx = s_idx
                    assert schema_idx is not None
                    # find the schema in ref_command
                    ref_s = cls.find_schema_in_command(ref_command, schema_idx)
                    if not ref_s:
                        continue
                    if not isinstance(ref_s, (CMDObjectSchemaBase, CMDArraySchemaBase)):
                        continue
                    if ref_s.cls:
                        continue
                    unwrap_arg_vars.add(arg_var)

                for arg_var in unwrap_arg_vars:
                    cls._unwrap_cls_arg_in_command(command, arg_var)

                # When unwrapped, the schema changed, so it's required to continue to verify the cls_name
                unwrap_detect = len(unwrap_arg_vars) > 0

        # inherit arguments modification
        ref_args = []
        if ref_command.arg_groups:
            for group in ref_command.arg_groups:
                ref_args.extend(group.args)
        command.generate_args(ref_args=ref_args)

        # inherit outputs
        command.generate_outputs(ref_outputs=ref_command.outputs)

    def remove_subresource_commands(self, resource_id, version, subresource):
        commands = []

        cmd_names_list = [cmd_names for cmd_names, _ in self.iter_commands_by_resource(resource_id, subresource, version)]
        if not cmd_names_list:
            return commands

        for cmd_names in cmd_names_list:
            commands.append(self._remove_command(*cmd_names))

        self.reformat()
        return commands

    def build_subresource_commands_by_arg_var(self, resource_id, arg_var, cg_names, ref_args_options=None):
        """

        :param resource_id:
        :param arg_var:
        :param cg_names: command group name for subresource commands
        :param ref_args_options: reference argument options
        :return:
        """
        if isinstance(cg_names, str):
            cg_names = [w for w in cg_names.split(' ') if w]

        assert isinstance(cg_names, list)

        update_cmd_info = self.get_update_cmd(resource_id)
        if not update_cmd_info:
            raise exceptions.InvalidAPIUsage(f"Resource does not exist generic update command: resource_id={resource_id}")

        update_cmd_names, update_cmd, update_by = update_cmd_info
        if update_by != "GenericOnly":
            raise exceptions.InvalidAPIUsage(f"Resource does not exist generic update command: resource_id={resource_id}")

        if arg_var.startswith('@'):
            raise exceptions.InvalidAPIUsage(f"Not support class arg={arg_var}, please unwrap it first.")

        update_op = None
        for operation in update_cmd.operations:
            if isinstance(operation, CMDInstanceUpdateOperation):
                update_op = operation
        assert update_op is not None

        # find schema_idx
        schema = None
        schema_idx = None
        for _, s, s_idx in self.iter_schema_in_operation_by_arg_var(update_op, arg_var):
            assert schema is None
            schema, schema_idx = s, s_idx
        if not schema:
            raise exceptions.InvalidAPIUsage(f"Cannot find schema for arg_var={arg_var}")

        subresource_idx = self.schema_idx_to_subresource_idx(schema_idx)
        assert subresource_idx

        sub_commands = self._generate_sub_commands(schema, subresource_idx, update_cmd, ref_args_options)

        for sub_command in sub_commands:
            self._add_command(*cg_names, sub_command.name, command=sub_command)

        self.reformat()

    @classmethod
    def _generate_sub_commands(cls, schema, subresource_idx, update_cmd, ref_args_options=None):
        update_op = None
        for operation in update_cmd.operations:
            if isinstance(operation, CMDInstanceUpdateOperation):
                update_op = operation
        assert update_op is not None

        # find arg
        if schema.arg:
            arg_var = schema.arg
            arg, arg_idx = cls.find_arg_in_command_by_var(update_cmd, arg_var)
            if not arg or not arg_idx:
                raise exceptions.InvalidAPIUsage(f"Argument '{arg_var}' not exist in command")
        else:
            # schema is flatten
            arg = None

        # build ref_args
        ref_args = []
        update_json = update_op.instance_update.json
        update_arg_prefix = f"${update_json.schema.name}"
        if update_cmd.arg_groups:
            for g in update_cmd.arg_groups:
                for a in g.args:
                    if a.var.startswith(update_arg_prefix):
                        # ignore arguments linked with body
                        continue
                    if 'name' in a.options and isinstance(a, CMDStringArgBase):
                        # remove auto add 'name', 'n' options
                        a = a.__class__(raw_data=a.to_native())
                        a.options = sorted(a.options, key=lambda o: (len(o), o))[-1:]  # use the longest argument
                    ref_args.append(a)

        if isinstance(schema, CMDArraySchema):
            assert isinstance(arg, CMDArrayArg)
            # list command
            list_command = cls._build_subresource_list_or_show_command(
                update_cmd=update_cmd,
                subresource_idx=subresource_idx,
                ref_args=ref_args,
                ref_options=ref_args_options
            )

            # generate item command
            item_subresource_idx = [*subresource_idx, '[]']

            item_arg = arg.item
            if isinstance(item_arg, CMDClsArgBase):
                raise exceptions.InvalidAPIUsage(f"Not support array<class> arg={arg_var}, please unwrap it first.")

            if not isinstance(item_arg, CMDObjectArgBase):
                raise NotImplementedError()

            if item_arg.additional_props:
                raise NotImplementedError()

            ref_args.extend(item_arg.args)

            # create command
            create_command = cls._build_subresource_create_command(
                update_cmd=update_cmd,
                subresource_idx=item_subresource_idx,
                ref_args=ref_args,
                ref_options=ref_args_options,
            )
            # update command
            update_command = cls._build_subresource_update_command(
                update_cmd=update_cmd,
                subresource_idx=item_subresource_idx,
                ref_args=ref_args,
                ref_options=ref_args_options,
            )
            # delete command
            delete_command = cls._build_subresource_delete_command(
                update_cmd=update_cmd,
                subresource_idx=item_subresource_idx,
                ref_args=ref_args,
                ref_options=ref_args_options,
            )
            # show command
            show_command = cls._build_subresource_list_or_show_command(
                update_cmd=update_cmd,
                subresource_idx=item_subresource_idx,
                ref_args=ref_args,
                ref_options=ref_args_options,
            )

            list_command.name = 'list'
            create_command.name = 'create'
            update_command.name = 'update'
            delete_command.name = 'delete'
            show_command.name = 'show'

            return [list_command, create_command, update_command, delete_command, show_command]

        elif isinstance(schema, CMDObjectSchema):
            if schema.props and schema.additional_props:
                raise NotImplementedError()

            if schema.props:
                if arg:
                    assert isinstance(arg, CMDObjectArg)
                    ref_args.extend(arg.args)

                # create command
                create_command = cls._build_subresource_create_command(
                    update_cmd=update_cmd,
                    subresource_idx=subresource_idx,
                    ref_args=ref_args,
                    ref_options=ref_args_options
                )
                # update command
                update_command = cls._build_subresource_update_command(
                    update_cmd=update_cmd,
                    subresource_idx=subresource_idx,
                    ref_args=ref_args,
                    ref_options=ref_args_options
                )
                # delete command
                delete_command = cls._build_subresource_delete_command(
                    update_cmd=update_cmd,
                    subresource_idx=subresource_idx,
                    ref_args=ref_args,
                    ref_options=ref_args_options
                )
                # show command
                show_command = cls._build_subresource_list_or_show_command(
                    update_cmd=update_cmd,
                    subresource_idx=subresource_idx,
                    ref_args=ref_args,
                    ref_options=ref_args_options
                )

            elif schema.additional_props and schema.additional_props.item:
                assert isinstance(arg, CMDObjectArg)
                item_subresource_idx = [*subresource_idx, '{}']

                item_arg = arg.additional_props.item
                if isinstance(item_arg, CMDClsArgBase):
                    raise exceptions.InvalidAPIUsage(f"Not support dict<class> arg={arg_var}, please unwrap it first.")

                if not isinstance(item_arg, CMDObjectSchemaBase):
                    raise NotImplementedError()

                if item_arg.additional_props:
                    raise NotImplementedError()

                ref_args.extend(item_arg.args)

                # create command
                create_command = cls._build_subresource_create_command(
                    update_cmd=update_cmd,
                    subresource_idx=item_subresource_idx,
                    ref_args=ref_args,
                    ref_options=ref_args_options
                )
                # update command
                update_command = cls._build_subresource_update_command(
                    update_cmd=update_cmd,
                    subresource_idx=item_subresource_idx,
                    ref_args=ref_args,
                    ref_options=ref_args_options
                )
                # delete command
                delete_command = cls._build_subresource_delete_command(
                    update_cmd=update_cmd,
                    subresource_idx=item_subresource_idx,
                    ref_args=ref_args,
                    ref_options=ref_args_options
                )
                # show command
                show_command = cls._build_subresource_list_or_show_command(
                    update_cmd=update_cmd,
                    subresource_idx=item_subresource_idx,
                    ref_args=ref_args,
                    ref_options=ref_args_options
                )
            else:
                raise NotImplementedError()

            create_command.name = 'create'
            update_command.name = 'update'
            delete_command.name = 'delete'
            show_command.name = 'show'

            return [create_command, update_command, delete_command, show_command]
        else:
            raise NotImplementedError()

    @classmethod
    def _build_sub_command_base(cls, update_cmd, subresource_idx, **kwargs):
        get_op = None
        put_op = None
        update_op = None
        for operation in update_cmd.operations:
            if isinstance(operation, CMDInstanceUpdateOperation):
                update_op = operation
            if isinstance(operation, CMDHttpOperation) and operation.http.request and operation.http.responses:
                if operation.http.request.method == "get":
                    get_op = operation
                if operation.http.request.method == "put":
                    put_op = operation
        assert get_op is not None
        assert put_op is not None
        assert update_op is not None

        # find response body
        response_json = None
        for response in get_op.http.responses:
            if response.is_error:
                continue
            if not isinstance(response.body, CMDHttpResponseJsonBody):
                continue
            if response.body.json.var == update_op.instance_update.ref:
                response_json = response.body.json
                break
        assert response_json is not None

        update_json = update_op.instance_update.json

        _sub_command = CMDCommand()
        _sub_command.version = update_cmd.version
        assert len(update_cmd.resources) == 1
        _resource = CMDResource(raw_data=update_cmd.resources[0].to_native())
        _resource.subresource = cls.idx_to_str(subresource_idx)
        _sub_command.resources = [_resource]
        _sub_command.subresource_selector = cls._build_subresource_selector(
            response_json, update_json, subresource_idx, **kwargs)
        return _sub_command, get_op, put_op, update_json

    @classmethod
    def _build_subresource_list_or_show_command(cls, update_cmd, subresource_idx, ref_args, ref_options):
        _sub_command, get_op, _, update_json = cls._build_sub_command_base(update_cmd, subresource_idx)

        _sub_command.operations = [get_op.__class__(raw_data=get_op.to_native())]
        _sub_command.generate_args(ref_args=ref_args, ref_options=ref_options)
        _sub_command.generate_outputs(ref_outputs=update_cmd.outputs)
        _sub_command.link()
        return _sub_command

    @classmethod
    def _build_subresource_create_command(cls, update_cmd, subresource_idx, ref_args, ref_options):
        _sub_command, get_op, put_op, update_json = cls._build_sub_command_base(update_cmd, subresource_idx, optional_end_index_indentifier=True)

        _instance_op = CMDInstanceCreateOperation()
        _instance_op.instance_create = CMDJsonInstanceCreateAction()
        _instance_op.instance_create.ref = _sub_command.subresource_selector.var
        _instance_op.instance_create.json = CMDRequestJson()

        # fork json schema
        assert update_cmd.schema_cls_register_map is not None
        _instance_op_schema = cls.fork_schema_in_json(
            update_json, subresource_idx, update_cmd.schema_cls_register_map)
        assert not isinstance(_instance_op_schema, CMDClsSchemaBase)
        if not isinstance(_instance_op_schema, CMDSchema):
            # convert schema base to schema
            if isinstance(_instance_op_schema, CMDObjectSchemaBase):
                _instance_op_schema = CMDObjectSchema(raw_data=_instance_op_schema.to_native())
            elif isinstance(_instance_op_schema, CMDArraySchemaBase):
                _instance_op_schema = CMDArraySchema(raw_data=_instance_op_schema.to_native())
            else:
                raise NotImplementedError()
        # _subresource_idx does not contains update_json.schema.name
        _instance_op_schema.name = cls.idx_to_str([update_json.schema.name, *subresource_idx])
        _instance_op_schema.required = True
        _instance_op.instance_create.json.schema = _instance_op_schema

        _sub_command.operations = [
            get_op.__class__(raw_data=get_op.to_native()),
            _instance_op,
            put_op.__class__(raw_data=put_op.to_native()),
        ]
        _sub_command.generate_args(ref_args=ref_args, ref_options=ref_options)
        _sub_command.generate_outputs(ref_outputs=update_cmd.outputs)
        _sub_command.link()
        return _sub_command

    @classmethod
    def _build_subresource_update_command(cls, update_cmd, subresource_idx, ref_args, ref_options):
        _sub_command, get_op, put_op, update_json = cls._build_sub_command_base(update_cmd, subresource_idx)

        _instance_op = CMDInstanceUpdateOperation()
        _instance_op.instance_update = CMDJsonInstanceUpdateAction()
        _instance_op.instance_update.ref = _sub_command.subresource_selector.var
        _instance_op.instance_update.json = CMDRequestJson()

        # fork json schema
        assert update_cmd.schema_cls_register_map is not None
        _instance_op_schema = cls.fork_schema_in_json(
            update_json, subresource_idx, update_cmd.schema_cls_register_map)
        if not isinstance(_instance_op_schema, CMDSchema):
            # convert schema base to schema
            if isinstance(_instance_op_schema, CMDObjectSchemaBase):
                _instance_op_schema = CMDObjectSchema(raw_data=_instance_op_schema.to_native())
            elif isinstance(_instance_op_schema, CMDArraySchemaBase):
                _instance_op_schema = CMDArraySchema(raw_data=_instance_op_schema.to_native())
            else:
                raise NotImplementedError()
        # _subresource_idx does not contains update_json.schema.name
        _instance_op_schema.name = cls.idx_to_str([update_json.schema.name, *subresource_idx])
        _instance_op_schema.required = True
        _instance_op.instance_update.json.schema = _instance_op_schema

        _sub_command.operations = [
            get_op.__class__(raw_data=get_op.to_native()),
            _instance_op,
            put_op.__class__(raw_data=put_op.to_native()),
        ]
        _sub_command.generate_args(ref_args=ref_args, ref_options=ref_options)
        _sub_command.generate_outputs(ref_outputs=update_cmd.outputs)
        _sub_command.link()
        return _sub_command

    @classmethod
    def _build_subresource_delete_command(cls, update_cmd, subresource_idx, ref_args, ref_options):
        _sub_command, get_op, put_op, _ = cls._build_sub_command_base(update_cmd, subresource_idx)

        _instance_op = CMDInstanceDeleteOperation()
        _instance_op.instance_delete = CMDJsonInstanceDeleteAction()
        _instance_op.instance_delete.ref = _sub_command.subresource_selector.var
        _instance_op.instance_delete.json = CMDRequestJson()
        _sub_command.operations = [
            get_op.__class__(raw_data=get_op.to_native()),
            _instance_op,
            put_op.__class__(raw_data=put_op.to_native()),
        ]
        _sub_command.confirmation = DEFAULT_CONFIRMATION_PROMPT
        _sub_command.generate_args(ref_args=ref_args, ref_options=ref_options)
        _sub_command.generate_outputs(ref_outputs=update_cmd.outputs)
        _sub_command.link()
        return _sub_command

    @classmethod
    def _build_subresource_selector(cls, response_json, update_json, subresource_idx, **kwargs):
        assert isinstance(response_json, CMDResponseJson)
        assert isinstance(update_json, CMDRequestJson)

        idx = cls.idx_to_list(subresource_idx)
        selector = CMDJsonSubresourceSelector()
        selector.ref = response_json.var

        if isinstance(update_json.schema, CMDObjectSchema):
            assert isinstance(response_json.schema, CMDObjectSchemaBase)
            index = CMDObjectIndex()
            index.name = update_json.schema.name
            # disable prune for the outer layer index
            selector.json = cls._build_object_index_base(update_json.schema, idx, index=index, prune=False, **kwargs)
        elif isinstance(update_json.schema, CMDArraySchema):
            assert isinstance(response_json.schema, CMDArraySchemaBase)
            index = CMDArrayIndex()
            index.name = update_json.schema.name
            selector.json = cls._build_array_index_base(update_json.schema, idx, index=index, **kwargs)
        else:
            raise NotImplementedError(f"Not support schema {type(update_json.schema)}")

        return selector

    @classmethod
    def _build_simple_index_base(cls, schema, idx, index=None, **kwargs):
        if index is None:
            if not isinstance(schema, CMDSimpleIndexBase.supported_schema_types):
                raise NotImplementedError(f"Not support schema '{type(schema)}'")
            index = CMDSimpleIndexBase()
        if idx:
            raise exceptions.InvalidAPIUsage(f"Not support remain index {idx}")
        return index

    @classmethod
    def _build_object_index_base(cls, schema, idx, index=None, prune=False, **kwargs):
        assert isinstance(schema, CMDObjectSchemaBase)

        if not index:
            index = CMDObjectIndexBase()

        if not idx:
            return index

        current_idx = idx[0]
        remain_idx = idx[1:]
        find_idx = False

        if schema.props:
            for prop in schema.props:
                if prop.name != current_idx:
                    continue
                find_idx = True
                if prune:
                    assert isinstance(index, CMDSelectorIndex)
                    name = f"{index.name}.{prop.name}"
                    if isinstance(prop, CMDClsSchema):
                        prop = prop.implement
                    # ignore the current index, return the sub index with index.name prefix
                    if isinstance(prop, CMDObjectSchema):
                        return cls._build_object_index(prop, remain_idx, name=name, **kwargs)
                    elif isinstance(prop, CMDArraySchema):
                        return cls._build_array_index(prop, remain_idx, name=name, **kwargs)
                    else:
                        return cls._build_simple_index(prop, remain_idx, name=name, **kwargs)
                else:
                    name = prop.name
                    if isinstance(prop, CMDClsSchema):
                        prop = prop.implement
                    if isinstance(prop, CMDObjectSchema):
                        index.prop = cls._build_object_index(prop, remain_idx, name=name, **kwargs)
                        break
                    elif isinstance(prop, CMDArraySchema):
                        index.prop = cls._build_array_index(prop, remain_idx, name=name, **kwargs)
                        break
                    else:
                        index.prop = cls._build_simple_index(prop, remain_idx, name=name, **kwargs)
                        break

        if schema.discriminators:
            for disc in schema.discriminators:
                if disc.get_safe_value() == current_idx:
                    find_idx = True
                    index.discriminator = cls._build_object_index_discriminator(disc, remain_idx, **kwargs)
                    break

        if schema.additional_props and current_idx == "{}":
            find_idx = True
            index.additional_props = cls._build_object_index_additional_prop(schema.additional_props, remain_idx, **kwargs)

        if not find_idx:
            raise exceptions.InvalidAPIUsage(f"Cannot find remain index {idx}")

        return index

    @classmethod
    def _build_array_index_base(cls, schema, idx, index=None, **kwargs):
        assert isinstance(schema, CMDArraySchemaBase)

        if not index:
            index = CMDArrayIndexBase()

        if not idx:
            return index

        current_idx = idx[0]
        remain_idx = idx[1:]
        if current_idx != '[]':
            raise exceptions.InvalidAPIUsage(f"Cannot find index '{idx}'")

        item = schema.item
        if isinstance(item, CMDClsSchemaBase):
            item = item.implement

        identifiers = []
        identifier_names = schema.identifiers
        if not identifier_names and isinstance(item, CMDObjectSchemaBase):
            prop_names = {prop.name for prop in item.props}
            if 'id' in prop_names and 'name' in prop_names:
                # use name as default identifier when schema contains 'id' property
                identifier_names = ['name']

        if identifier_names:
            assert isinstance(item, CMDObjectSchemaBase)
            for prop in item.props:
                if prop.name in identifier_names:
                    identifier = prop.__class__(raw_data=prop.to_native())
                    identifier.name = '[].' + prop.name
                    identifier.required = True
                    identifier.read_only = False
                    identifier.frozen = False
                    identifiers.append(identifier)

        if not identifiers:
            # use index as identifier
            identifier = CMDIntegerSchema()
            identifier.name = '[Index]'
            identifiers.append(identifier)
            identifier.required = len(remain_idx) > 0 or not kwargs.get('optional_end_index_indentifier', False)
        index.identifiers = identifiers

        if isinstance(item, CMDObjectSchemaBase):
            index.item = cls._build_object_index_base(item, remain_idx, **kwargs)
        elif isinstance(item, CMDArraySchemaBase):
            index.item = cls._build_array_index_base(item, remain_idx, **kwargs)
        else:
            index.item = cls._build_simple_index_base(item, remain_idx, **kwargs)

        return index

    @classmethod
    def _build_simple_index(cls, schema, idx, name, **kwargs):
        if not isinstance(schema, CMDSimpleIndex.supported_schema_types):
            raise NotImplementedError(f"Not support schema '{type(schema)}'")
        index = CMDSimpleIndex()
        index.name = name
        return cls._build_simple_index_base(schema, idx, index=index, **kwargs)

    @classmethod
    def _build_object_index(cls, schema, idx, name, **kwargs):
        index = CMDObjectIndex()
        index.name = name
        return cls._build_object_index_base(schema, idx, index, prune=True, **kwargs)

    @classmethod
    def _build_array_index(cls, schema, idx, name, **kwargs):
        index = CMDArrayIndex()
        index.name = name
        return cls._build_array_index_base(schema, idx, index, **kwargs)

    @classmethod
    def _build_object_index_discriminator(cls, schema, idx, **kwargs):
        assert isinstance(schema, CMDObjectSchemaDiscriminator)

        index = CMDObjectIndexDiscriminator()
        index.property = schema.property
        index.value = schema.value

        if not idx:
            return index

        current_idx = idx[0]
        remain_idx = idx[1:]

        find_idx = False

        if schema.props:
            for prop in schema.props:
                if prop.name != current_idx:
                    continue
                find_idx = True
                name = prop.name
                if isinstance(prop, CMDClsSchema):
                    prop = prop.implement

                if isinstance(prop, CMDObjectSchema):
                    index.prop = cls._build_object_index(prop, remain_idx, name, **kwargs)
                    break
                elif isinstance(prop, CMDArraySchema):
                    index.prop = cls._build_array_index(prop, remain_idx, name, **kwargs)
                    break
                else:
                    index.prop = cls._build_simple_index(prop, remain_idx, name, **kwargs)
                    break

        if schema.discriminators:
            for disc in schema.discriminators:
                if disc.get_safe_value() == current_idx:
                    find_idx = True
                    index.discriminator = cls._build_object_index_discriminator(disc, remain_idx, **kwargs)
                    break

        if not find_idx:
            raise exceptions.InvalidAPIUsage(f"Cannot find remain index {idx}")

        return index

    @classmethod
    def _build_object_index_additional_prop(cls, schema, idx, **kwargs):
        assert isinstance(schema, CMDObjectSchemaAdditionalProperties)
        index = CMDObjectIndexAdditionalProperties()

        identifier = CMDStringSchema()
        identifier.name = "{Key}"
        identifier.required = True
        index.identifiers = [identifier]

        item = schema.item
        if isinstance(item, CMDClsSchemaBase):
            item = item.implement
        if isinstance(item, CMDObjectSchemaBase):
            index.item = cls._build_object_index_base(item, idx, **kwargs)
        elif isinstance(item, CMDArraySchemaBase):
            index.item = cls._build_array_index_base(item, idx, **kwargs)
        else:
            index.item = cls._build_simple_index_base(item, idx, **kwargs)

        return index

    @classmethod
    def fork_schema_in_json(cls, js, idx, reference_cls_map):
        from command.model.configuration._content import _iter_over_schema_for_cls_register
        schema = cls.find_schema_in_json(js, idx)
        if not schema:
            return None

        if isinstance(schema, CMDClsSchemaBase):
            assert schema.implement is not None
            schema = schema.get_unwrapped()
        else:
            schema = schema.__class__(raw_data=schema.to_native())
        assert not isinstance(schema, CMDClsSchemaBase)

        # make sure cls implement contained in schema
        cls_register_map = {}
        _iter_over_schema_for_cls_register(schema, cls_register_map)
        miss_implement_cls_names = set()
        for k, v in cls_register_map.items():
            if not v['implement']:
                miss_implement_cls_names.add(k)

        pre_miss_implement_cls_names = set()
        while len(miss_implement_cls_names) > 0 and pre_miss_implement_cls_names != miss_implement_cls_names:
            for cls_name in miss_implement_cls_names:
                # unwrap the miss implement clsSchema to instance
                for ref_parent, ref_schema, _ in cls.iter_schema_cls_reference_in_schema(schema, cls_name):
                    assert isinstance(ref_schema, CMDClsSchemaBase)
                    assert cls_register_map[cls_name]['implement'] is None
                    ref_schema.implement = reference_cls_map[cls_name]['implement']
                    assert ref_schema.implement is not None
                    new_ref_schema = ref_schema.get_unwrapped()
                    assert new_ref_schema.cls == cls_name
                    cls.replace_schema(ref_parent, ref_schema, new_ref_schema)
                    break

            # recalculate miss implement cls
            pre_miss_implement_cls_names = miss_implement_cls_names
            cls_register_map = {}
            _iter_over_schema_for_cls_register(schema, cls_register_map)
            miss_implement_cls_names = set()
            for k, v in cls_register_map.items():
                if not v['implement']:
                    miss_implement_cls_names.add(k)

        if len(miss_implement_cls_names) > 0:
            raise exceptions.InvalidAPIUsage(f"Always miss class implements for {miss_implement_cls_names}")

        return schema


def build_endpoint_selector_for_client_config(get_op, subresource_idx, **kwargs):
    # find response body
    response_json = None
    for response in get_op.http.responses:
        if response.is_error:
            continue
        if not isinstance(response.body, CMDHttpResponseJsonBody):
            continue
        if response.body.json.var == CMDBuildInVariants.EndpointInstance:
            response_json = response.body.json
            break

    assert isinstance(response_json, CMDResponseJson)
    idx = WorkspaceCfgEditor.idx_to_list(subresource_idx)
    selector = CMDJsonSubresourceSelector()
    selector.ref = response_json.var
    selector.var = CMDBuildInVariants.Endpoint
    if isinstance(response_json.schema, CMDObjectSchemaBase):
        index = CMDObjectIndex()
        index.name = 'response'
        selector.json = WorkspaceCfgEditor._build_object_index_base(
            response_json.schema, idx, index=index, **kwargs)
    elif isinstance(response_json.schema, CMDArraySchemaBase):
        index = CMDArrayIndex()
        index.name = 'response'
        selector.json = WorkspaceCfgEditor._build_array_index_base(
            response_json.schema, idx, index=index, **kwargs)
    else:
        raise NotImplementedError(f"Not support schema {type(response_json.schema)}")

    return selector



from cli.model.common._fields import CLIResourceIdField, CLIVersionField
from schematics.models import Model
from schematics.types import StringType
from utils.fields import PlaneField


class CLISpecsResource(Model):
    plane = PlaneField(required=True)
    id = CLIResourceIdField(required=True)
    version = CLIVersionField(required=True)
    subresource = StringType()  # subresource index, used for sub commands generation.

    class Options:
        serialize_when_none = False



from schematics.models import Model
from schematics.types import StringType, URLType


class Contact(Model):
    """Contact information for the exposed API."""

    name = StringType()    # The identifying name of the contact person/organization.
    url = URLType()        # The URL pointing to the contact information. MUST be in the format of a URL.
    email = StringType()   # The email address of the contact person/organization. MUST be in the format of an email address.

    class Options:
        serialize_when_none = False



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
import re

URL_PARAMETER_PLACEHOLDER = "{}"


def swagger_resource_path_to_resource_id_template(path):
    path_parts = path.split("?", maxsplit=1)
    url_parts = path_parts[0].split("/")
    idx = 1
    while idx < len(url_parts):
        if idx == 1 and re.match(r'^\{[^{}]*}$', url_parts[idx]):
            # ignore to the parameter name in first part of url, such as `/{parameter}` or `/{parameter}/...`
            # But the placeholder in first part like `/indexes('{indexname}')` will be processed
            idx += 1
            continue
        url_parts[idx] = re.sub(r'\{[^{}]*}', URL_PARAMETER_PLACEHOLDER, url_parts[idx])
        idx += 1
    path_parts[0] = "/".join(url_parts)
    return "?".join(path_parts)


def swagger_resource_path_to_resource_id(path):
    path_parts = path.split("?", maxsplit=1)
    url_parts = path_parts[0].split("/")
    idx = 1
    while idx < len(url_parts):
        if idx == 1 and re.match(r'^\{[^{}]*}$', url_parts[idx]):
            # ignore to the parameter name in first part of url, such as `/{parameter}` or `/{parameter}/...`
            # But the placeholder in first part like `/indexes('{indexname}')` will be processed
            idx += 1
            continue
        # replace all placeholders in part
        url_parts[idx] = re.sub(r'\{[^{}]*}', URL_PARAMETER_PLACEHOLDER, url_parts[idx])
        idx += 1
    path_parts[0] = "/".join(url_parts).lower()
    return "?".join(path_parts)



import datetime
import json
import logging
import os
import re
from collections import OrderedDict

import yaml

from swagger.utils.tools import swagger_resource_path_to_resource_id
from ._resource import Resource, ResourceVersion
from ._utils import map_path_2_repo

logger = logging.getLogger('backend')


class OpenAPIResourceProvider:

    def __init__(self, name, folder_path, readme_path, swagger_module):
        self.name = name
        self.folder_path = folder_path
        self._readme_path = readme_path
        self.swagger_module = swagger_module

        if readme_path is None:
            logger.warning(f"MissReadmeFile: {self} : {map_path_2_repo(folder_path)}")
        self._tags = None
        self._resource_map = None
        self._ignore_resources = {f'/providers/{self.name}/operations'.lower(), }

    def __str__(self):
        return f'{self.swagger_module}/ResourceProviders/{self.name}'

    def get_resource_map(self, refresh=False):
        if refresh or not self._resource_map:
            resource_map = {}
            for root, dirs, files in os.walk(self.folder_path):
                if 'example' in root:
                    continue
                for file in files:
                    if not file.endswith('.json'):
                        continue
                    file_path = os.path.join(root, file)
                    for resource in self._parse_resources_in_file(file_path):
                        if resource.id in self._ignore_resources:
                            continue
                        if resource.id not in resource_map:
                            resource_map[resource.id] = {}
                        if self._replace_current_resource(
                                curr_resource=resource_map[resource.id].get(resource.version, None),
                                resource=resource
                        ):
                            resource_map[resource.id][resource.version] = resource
            self._resource_map = resource_map
        resource_map = self._resource_map
        return resource_map

    def get_resource_map_by_tag(self, tag):
        resource_map = {}
        if tag not in self.tags:
            logger.error(f"Tag: `{tag}` is not exist")
            return resource_map

        for file_path in self.tags[tag]:
            for resource in self._parse_resources_in_file(file_path):
                if resource.id in self._ignore_resources:
                    continue
                if resource.id not in resource_map:
                    resource_map[resource.id] = {}
                if self._replace_current_resource(
                        curr_resource=resource_map[resource.id].get(resource.version, None),
                        resource=resource
                ):
                    resource_map[resource.id][resource.version] = resource
        return resource_map

    @property
    def tags(self):
        if self._tags is None:
            self._tags = self._parse_readme_input_file_tags()
        return self._tags

    def _parse_readme_input_file_tags(self):
        tags = {}
        if self._readme_path is None:
            return tags

        with open(self._readme_path, 'r', encoding='utf-8') as f:
            readme = f.read()

        re_yaml = re.compile(
            r'```\s*yaml\s*(.*\$\(\s*tag\s*\)\s*==\s*[\'"]\s*(.*)\s*[\'"].*)?\n((((?!```).)*\n)*)```\s*\n',
            flags=re.MULTILINE)
        for piece in re_yaml.finditer(readme):
            flags = piece[1]
            yaml_body = piece[3]
            if 'input-file' not in yaml_body:
                continue

            try:
                body = yaml.safe_load(yaml_body)
            except yaml.YAMLError as err:
                logger.error(f'ParseYamlFailed: {self} : {self._readme_path} {flags}: {err}')
                continue

            files = []
            for file_path in body['input-file']:
                file_path = file_path.replace('$(this-folder)/', '')
                file_path = os.path.join(os.path.dirname(self._readme_path), *file_path.split('/'))
                if not os.path.isfile(file_path):
                    logger.warning(f'FileNotExist: {self} : {file_path}')
                    continue
                files.append(file_path)

            if len(files):
                tag = piece[2]
                if tag is None:
                    tag = ''
                tag = OpenAPIResourceProviderTag(tag.strip(), self)
                if tag not in tags:
                    tags[tag] = set()
                tags[tag] = tags[tag].union(files)

        tags = [*tags.items()]
        tags.sort(key=lambda item: item[0].date, reverse=True)
        tags = OrderedDict(tags)
        return tags

    def _fetch_latest_tag(self, file_path):
        for tag, file_set in self.tags.items():
            if file_path in file_set:
                return tag
        return None

    def _replace_current_resource(self, curr_resource, resource):
        if curr_resource is None:
            # previous resource with same path and version
            return True

        assert curr_resource.id == resource.id and curr_resource.version == resource.version

        if curr_resource.path != resource.path:
            logger.warning(
                f'SimilarResourcePath: {self} :\n'
                f'\tFile: {map_path_2_repo(curr_resource.file_path)} Path: {curr_resource.path}\n'
                f'\tFile: {map_path_2_repo(resource.file_path)} Path: {resource.path}')

        curr_rs_tag = self._fetch_latest_tag(curr_resource.file_path)
        rs_tag = self._fetch_latest_tag(resource.file_path)

        # resource's file path used in tag has higher priority
        if curr_rs_tag is None and rs_tag is not None:
            return True
        elif curr_rs_tag is not None and rs_tag is None:
            return False

        # resource's file path with larger date has higher priority
        if curr_resource.file_path_version.date is not None and resource.file_path_version.date is not None:
            if curr_resource.file_path_version.date < resource.file_path_version.date:
                return True
            elif curr_resource.file_path_version.date > resource.file_path_version.date:
                return False

        # resource's file path in stable has higher priority
        if curr_resource.file_path_version.readiness != resource.file_path_version.readiness:
            if resource.file_path_version.readiness == ResourceVersion.Readiness.Stable:
                return True
            elif curr_resource.file_path_version == ResourceVersion.Readiness.Stable:
                return False

        # resource's file in latest tag has higher priority
        if curr_rs_tag is not None and rs_tag is not None:
            if rs_tag.date > curr_rs_tag.date:
                return True
            elif curr_rs_tag.date > rs_tag.date:
                return False

        logger.warning(f'DuplicatedResource: {self} :\n'
                       f'\tFile: {map_path_2_repo(curr_resource.file_path)} Path: {curr_resource.path}\n'
                       f'\tFile: {map_path_2_repo(resource.file_path)} Path: {resource.path}')
        return False

    def _parse_resources_in_file(self, file_path):
        resources = []

        try:
            with open(file_path, 'r', encoding='utf-8') as f:
                body = json.load(f)
        except Exception as err:
            logger.error(f'InvalidSwaggerFile: {self} : ParseJsonFailed: {file_path} : {err}')
            return resources

        # check swagger version
        swagger_version = body.get('swagger', None)
        if swagger_version != '2.0':
            logger.error(f'InvalidSwaggerFile: {self} : invalid swagger version {swagger_version} in file {file_path}')
            return resources

        # fetch api-version
        info = body.get('info', {})
        version = info.get('version', None)
        if not version:
            logger.error(f'InvalidSwaggerFile: {self} : invalid info version {version} in file {file_path}')

        for path, value in body.get('paths', {}).items():
            resource = Resource(
                resource_id=swagger_resource_path_to_resource_id(path),
                path=path, version=version, file_path=file_path, resource_provider=self, body=value)
            resources.append(resource)

        # x-ms-paths:
        #   alternative to Paths Object that allows Path Item Object to have query parameters for non pure REST APIs
        for path, value in body.get('x-ms-paths', {}).items():
            resource = Resource(
                resource_id=swagger_resource_path_to_resource_id(path),
                path=path, version=version, file_path=file_path, resource_provider=self, body=value)
            resources.append(resource)

        return resources


class OpenAPIResourceProviderTag:

    def __init__(self, tag, resource_provider):
        self._tag = tag
        self.date = None

        if self._tag is not None:
            dt_re = re.compile(r'([0-9]{4})([-_]([0-9]{1,2})([-_]([0-9]{1,2}))?)?')
            try:
                pieces = next(dt_re.finditer(self._tag))
                year = int(pieces[1])
                month = int(pieces[3]) if pieces[3] else 1
                day = int(pieces[5]) if pieces[5] else 1
                self.date = datetime.date(year, month, day)
            except (StopIteration, ValueError):
                logger.warning(f'ParseTagDateError: {resource_provider} : {self._tag}')
                self.date = datetime.date.min

    def __str__(self):
        return self._tag

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)


class TypeSpecResourceProvider:

    def __init__(self, name, entry_files, swagger_module):
        self.name = name
        self.entry_files = entry_files
        self.swagger_module = swagger_module

    def __str__(self):
        return f'{self.swagger_module}/ResourceProviders/{self.name}'



def serialize(obj):
    def dfs(obj):
        if isinstance(obj, dict):
            return "{" + ",".join(f'{k}:{dfs(v)}' for k, v in obj.items()) + "}"

        elif isinstance(obj, list):
            return "[" + ",".join(dfs(i) for i in obj) + "]"

        else:
            if obj is None:
                return "null"  # replace None by null

            escaped = str(obj)
            escaped = escaped.replace("'", "'/")  # use '/ to input '

            # with space, null/help expressions or other special characters
            if any(char in escaped for char in (" ", "null", "??", ":", ",", "{", "}", "[", "]")):
                escaped = "'" + escaped + "'"

            return repr(escaped)[1:-1]  # ignore quotes

    return '"' + dfs(obj) + '"' if isinstance(obj, dict) or isinstance(obj, list) else obj





class InvalidSwaggerValueError(ValueError):

    def __init__(self, msg, key, value=None):
        super(InvalidSwaggerValueError, self).__init__()
        self.msg = msg
        self.key = key
        self.value = value

    def __str__(self):
        if self.value:
            return f"{self.__class__.__name__}: {self.msg} : {self.value} : {self.key}"
        else:
            return f"{self.__class__.__name__}: {self.msg} : {self.key}"



from schematics.models import Model
from schematics.types import DictType, ModelType

from command.controller.shorthand import serialize
from command.model.configuration import CMDCommandExample
from .reference import Linkable, ReferenceField


class ExampleItem(Model, Linkable):
    ref = ReferenceField()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ref_instance = None

    def link(self, swagger_loader, *traces):
        if self.is_linked():
            return

        super().link(swagger_loader, *traces)

        self.ref_instance, instance_traces = swagger_loader.load_ref(self.ref, *self.traces, "ref")

    def to_cmd(self, example_builder, cmd_name, **kwargs):
        if not self.ref_instance:
            return

        example_params = example_builder.mapping(self.ref_instance.get("parameters", {}))

        command = cmd_name
        for param_option, param_value in example_params:
            if param_option == "subscription":
                continue

            if len(param_option) == 1:
                command += f" -{param_option} {serialize(param_value)}"
            else:
                command += f" --{param_option} {serialize(param_value)}"

        return CMDCommandExample({"commands": [command.strip()]})


class XmsExamplesField(DictType):
    """
    Describes the format for specifying examples for request and response of an operation in an OpenAPI definition.
    It is a dictionary of different variations of the examples for a given operation.

    https://github.com/Azure/azure-rest-api-specs/blob/master/documentation/x-ms-examples.md
    """

    def __init__(self, **kwargs):
        super().__init__(
            field=ModelType(ExampleItem),
            serialized_name="x-ms-examples",
            deserialize_from="x-ms-examples",
            **kwargs
        )



from schematics.models import Model
from schematics.types import ModelType, DictType

from ._command_group import CLIAtomicCommandGroup
from ._client import CLIAtomicClient
from cli.model.common._fields import CLIProfileNameField

from utils.plane import PlaneEnum


class CLIAtomicProfile(Model):
    name = CLIProfileNameField(required=True)
    command_groups = DictType(
        field=ModelType(CLIAtomicCommandGroup),
        serialized_name="commandGroups",
        deserialize_from="commandGroups"
    )
    _clients = DictType(
        field=ModelType(CLIAtomicClient),
        serialized_name="clients",
        deserialize_from="clients",
    )

    class Options:
        serialize_when_none = False

    @property
    def profile_folder_name(self):
        profile_folder_name = self.name.lower().replace('-', '_')
        if profile_folder_name != "latest":
            # for rest profiles such as 2019-03-01-hybrid, the folder name starts with digit,
            # it's not a valid python package name.
            profile_folder_name = "profile_" + profile_folder_name
        return profile_folder_name

    def get_client(self, command):
        if not self._clients:
            return None
        plane = command.resources[0].plane
        name = PlaneEnum.http_client(plane)
        return self._clients.get(f'{plane}/{name}', None)

    def add_client(self, client):
        if not self._clients:
            self._clients = {}
        self._clients[f'{client.plane}/{client.name}'] = client

    @property
    def clients(self):
        return [*self._clients.values()]



import os
import logging
import re

logger = logging.getLogger('backend')


class TypeSpecHelper:

    @staticmethod
    def _iter_entry_files(folder):
        if not os.path.isdir(folder):
            raise ValueError(f"Path not exist: {folder}")
        ts_path = os.path.join(folder, "main.tsp")
        cfg_path = os.path.join(folder, "tspconfig.yaml")
        if os.path.isfile(ts_path) and os.path.isfile(cfg_path):
            yield ts_path, cfg_path
        for root, dirs, files in os.walk(folder):
            ts_path = os.path.join(root, "main.tsp")
            cfg_path = os.path.join(root, "tspconfig.yaml")
            if os.path.isfile(ts_path) and os.path.isfile(cfg_path):
                yield ts_path, cfg_path

    @classmethod
    def find_mgmt_plane_entry_files(cls, folder):
        files = []
        for ts_path, cfg_path in cls._iter_entry_files(folder):
            namespace, is_mgmt_plane = cls._parse_main_tsp(ts_path)
            if is_mgmt_plane:
                files.append((namespace, ts_path, cfg_path))
        return files

    @classmethod
    def find_data_plane_entry_files(cls, folder):
        files = []
        for ts_path, cfg_path in cls._iter_entry_files(folder):
            namespace, is_mgmt_plane = cls._parse_main_tsp(ts_path)
            if not namespace:
                continue
            if not is_mgmt_plane:
                files.append((namespace, ts_path, cfg_path))
        return files

    @classmethod
    def _parse_main_tsp(cls, path):
        with open(path, 'r', encoding='utf-8') as f:
            lines = f.readlines()
        is_mgmt_plane = False
        namespace = None
        for line in lines:
            if line.startswith("@armProviderNamespace"):
                is_mgmt_plane = True
            if line.startswith("namespace "):
                assert namespace is None
                namespace = re.match(r"^namespace\s+([A-Za-z0-9.]+)", line).group(1)
                # armProviderNamespace will always be appeared before namespace
                break
        if namespace is None:
            # logger.warning("Failed to parse main tsp file: %s namespace is not exist.", path)
            return None, None
        return namespace, is_mgmt_plane



from schematics.models import Model
from schematics.types import StringType

from ._fields import CMDResourceIdField, CMDVersionField
from ._utils import CMDDiffLevelEnum
from utils.base64 import b64decode_str


class CMDResource(Model):
    # properties as tags
    id = CMDResourceIdField(required=True)
    version = CMDVersionField(required=True)
    subresource = StringType()  # subresource index, used for sub commands generation.

    swagger = StringType(required=True)  # swagger resource provider, <plane>/<path:mod_names>/ResourceProviders/<rp_name>/Paths/<base64Encoded Path>/V/<base64Encoded version>

    class Options:
        serialize_when_none = False

    @property
    def plane(self):
        return self.swagger.split('/')[0]

    @property
    def mod_names(self):
        return self.swagger.split("/ResourceProviders/")[0].split('/')[1:]

    @property
    def rp_name(self):
        return self.swagger.split("/ResourceProviders/")[1].split('/')[0]

    @property
    def path(self):
        return b64decode_str(self.swagger.split("/Paths/")[1].split('/')[0])

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.id != old.id:
                diff["id"] = f"{old.id} != {self.id}"
            if self.version != old.version:
                diff["version"] = f"{old.version} != {self.version}"
            if self.subresource != old.subresource:
                diff["subresource"] = f"{old.subresource} != {self.subresource}"
            if self.plane != old.plane:
                diff["plane"] = f"{old.plane} != {self.plane}"
            if self.rp_name != old.rp_name:
                diff["rp_name"] = f"{old.rp_name} != {self.rp_name}"

        if level >= CMDDiffLevelEnum.Structure:
            if self.mod_names != old.mod_names:
                diff["mod_names"] = f"{old.mod_names} != {self.mod_names}"
            if self.path != old.path:
                diff["path"] = f"{old.path} != {self.path}"

        return diff



import inflect
import re

from lxml.builder import ElementMaker
from lxml.etree import Element, tostring
from xml.sax.saxutils import unescape
from xmltodict import parse

from schematics.types import ListType, ModelType
from schematics.types.compound import PolyModelType
from schematics.types.serializable import Serializable
from ._fields import CMDPrimitiveField

XML_ROOT = "CodeGen"


class XMLSerializer:

    @classmethod
    def to_xml(cls, value):
        primitive = value.to_primitive(context={"to_xml": True})
        root = build_xml(primitive)
        return unescape(
            tostring(root, xml_declaration=True, pretty_print=True, encoding="utf-8").decode()
        )

    @classmethod
    def from_xml(cls, model, xml):
        primitive = parse(cls._escape(xml), attr_prefix="")
        return build_model(model, primitive[XML_ROOT])

    @staticmethod
    def _escape(data):
        ret = ""
        lines = re.findall(r"<(.+)>", data)
        for line in lines:
            # handle long-summary
            if line.startswith("line>"):
                ret += "<" + line + ">" + "\n"
                continue
            line = line.replace("&", "&amp;")
            line = line.replace(">", "&gt;")
            line = line.replace("<", "&lt;")
            ret += "<" + line + ">" + "\n"
        return ret


def build_xml(primitive, parent=None):
    if parent is None:
        parent = getattr(ElementMaker(), XML_ROOT)()
    # normalize element name
    if elem_name := _inflect_engine.singular_noun(parent.tag):
        parent.tag = elem_name
    for field_name, data in primitive.items():
        primitive_to_xml(field_name, data, parent)
    return parent


def primitive_to_xml(field_name, data, parent):
    if isinstance(data, dict):
        _parent = getattr(ElementMaker(), field_name)()
        parent.append(build_xml(data, _parent))
    elif isinstance(data, list):
        for d in data:
            primitive_to_xml(field_name, d, parent)
    else:
        # handle long-summary
        if field_name == "line":
            child = Element("line")
            child.text = str(data)
            parent.append(child)
        # store metadata as attributes
        elif prev := parent.get(field_name):
            curr = " ".join(sorted(f"{prev} {data}".split(), key=len, reverse=True))
            parent.set(field_name, curr)
        else:
            parent.set(field_name, str(data))


def build_model(model, primitive):
    if hasattr(model, "_field_list"):
        instance = model()
        for field_name, field in model._field_list:
            if isinstance(field, Serializable):
                continue
            serialized_name = field.serialized_name or field_name
            # obtain suitable element name
            if serialized_name in primitive:
                curr_name = serialized_name
            elif (elem_name := _inflect_engine.singular_noun(serialized_name)) in primitive:
                curr_name = elem_name
            else:
                continue
            data = primitive[curr_name]
            curr_field = _unwrap(field)
            field_value = obtain_field_value(field, curr_field, data)
            setattr(instance, field_name, field_value)
        return instance
    else:
        # handle primitive field
        if model.primitive_type is not None:
            cast = model.primitive_type
        elif isinstance(model, CMDPrimitiveField):
            cast = model.convert_from_xml
        else:
            raise NotImplementedError(f"Cannot convert value from xml for field type: {type(model)}")
        return cast(primitive)


def obtain_field_value(prev, curr, data):
    if isinstance(prev, ListType):
        field_value = []
        # filter long-summary and singular options
        if " " in data and curr.parent_field.name not in {"lines", "singular_options"}:
            data = data.split()
        if isinstance(data, list):
            for d in data:
                model = curr.find_model(d) if isinstance(curr, PolyModelType) else curr
                value = build_model(model, d)
                field_value.append(value)
        else:
            model = curr.find_model(data) if isinstance(curr, PolyModelType) else curr
            value = build_model(model, data)
            field_value.append(value)
    else:
        model = curr.find_model(data) if isinstance(curr, PolyModelType) else curr
        field_value = build_model(model, data)
    return field_value


def _unwrap(field):
    if isinstance(field, ListType):
        return _unwrap(field.field)
    elif isinstance(field, ModelType):
        return field.model_class
    else:
        return field


_inflect_engine = inflect.engine()



from schematics.models import Model
from schematics.types import StringType, ListType, ModelType

from ._example import CLICommandExample


class CLICommandGroupHelp(Model):
    short = StringType(required=True)  # short-summary
    long = StringType()  # long-summary

    class Options:
        serialize_when_none = False


class CLICommandHelp(Model):
    short = StringType(required=True)  # short-summary
    long = StringType()  # long-summary
    examples = ListType(field=ModelType(CLICommandExample))

    class Options:
        serialize_when_none = False



from schematics.models import Model
from schematics.types import StringType, ModelType


class XmsPageable(Model):
    """
    The REST API guidelines define a common pattern for paging through lists of data. The operation response is modeled in OpenAPI as a list of items (a "page") and a link to the next page, effectively resembling a singly linked list. Tag the operation as x-ms-pageable and the generated code will include methods for navigating between pages.

    Note: The request to get the nextPage will always be a GET request

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-pageable
    """

    item_name = StringType(
        default='value',
        serialized_name="itemName",
        deserialize_from="itemName"
    )  # Optional (default: value). Specifies the name of the property that provides the collection of pageable items.
    next_link_name = StringType(
        required=True,
        serialized_name="nextLinkName",
        deserialize_from="nextLinkName"
    )  # Specifies the name of the property that provides the next link (common: nextLink). If the model does not have a next link property then specify null. This is useful for services that return an object that has an array referenced by itemName. The object is then flattened in a way that the array is directly returned, no paging is used. This provides a better client side API to the end user.
    operation_name = StringType(
        serialized_name="operationName",
        deserialize_from="operationName"
    )  # TODO: # Optional (default: <operationName>Next). Specifies the name of the operation for retrieving the next page.


class XmsPageableField(ModelType):

    def __init__(self, **kwargs):
        serialized_name=kwargs.pop('serialized_name', 'x-ms-pageable')
        deserialize_from=kwargs.pop('deserialize_from', 'x-ms-pageable')
        super(XmsPageableField, self).__init__(
            XmsPageable,
            serialized_name=serialized_name,
            deserialize_from=deserialize_from,
            **kwargs
        )



from schematics.models import Model
from schematics.types import ModelType, ListType, StringType

from .operation import TypeSpecOperation


class TypeSpecPathItem(Model):
    """Describes a single API operation on a path."""
    
    get = ModelType(TypeSpecOperation)  # A definition of a GET operation on this path.
    post = ModelType(TypeSpecOperation)  # A definition of a POST operation on this path.
    put = ModelType(TypeSpecOperation)  # A definition of a PUT operation on this path.
    patch = ModelType(TypeSpecOperation)  # A definition of a PATCH operation on this path.
    delete = ModelType(TypeSpecOperation)  # A definition of a DELETE operation on this path.
    head = ModelType(TypeSpecOperation)  # A definition of a HEAD operation on this path.

    traces = ListType(StringType())  # traces of the path item



from schematics.models import Model
from schematics.types import PolyModelType, ModelType, ListType, StringType

from ._fields import CMDVariantField, CMDDescriptionField
from ._http import CMDHttpAction
from ._instance_update import CMDInstanceUpdateAction
from ._instance_create import CMDInstanceCreateAction
from ._instance_delete import CMDInstanceDeleteAction
from ._utils import CMDDiffLevelEnum


class CMDOperation(Model):
    POLYMORPHIC_KEY = None

    # properties as tags
    when = ListType(CMDVariantField())  # conditions

    class Options:
        serialize_when_none = False

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.POLYMORPHIC_KEY is None:
            return False

        if isinstance(data, dict):
            return cls.POLYMORPHIC_KEY in data
        elif isinstance(data, CMDOperation):
            return hasattr(data, cls.POLYMORPHIC_KEY)
        return False

    def generate_args(self, ref_args, has_subresource, var_prefix=None):
        raise NotImplementedError()

    def reformat(self, **kwargs):
        raise NotImplementedError()

    def register_cls(self, **kwargs):
        return NotImplementedError()


class CMDHttpOperationLongRunning(Model):
    # properties as tags
    final_state_via = StringType(
        choices=(
            'azure-async-operation',
            # (default if not specified) poll until terminal state, the final response will be available at the uri pointed to by the header Azure-AsyncOperation
            'location',
            # poll until terminal state, the final response will be available at the uri pointed to by the header Location
            'original-uri'
            # poll until terminal state, the final response will be available via GET at the original resource URI. Very common for PUT operations.
        ),
        default='azure-async-operation',
        serialized_name='finalStateVia',
        deserialize_from='finalStateVia',
    )

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.final_state_via != old.final_state_via:
                diff["final_state_via"] = f"{old.final_state_via} != {self.final_state_via}"
        return diff


class CMDHttpOperation(CMDOperation):
    POLYMORPHIC_KEY = "http"

    # properties as tags
    long_running = ModelType(
        CMDHttpOperationLongRunning,
        serialized_name="longRunning",
        deserialize_from="longRunning",
    )

    operation_id = StringType(
        serialized_name="operationId",
        deserialize_from="operationId",
        required=True
    )   # OperationId from swagger

    description = CMDDescriptionField()

    # properties as nodes
    http = ModelType(CMDHttpAction, required=True)

    def generate_args(self, ref_args, has_subresource, var_prefix=None):
        return self.http.generate_args(ref_args=ref_args, has_subresource=has_subresource, var_prefix=var_prefix)

    def reformat(self, **kwargs):
        self.http.reformat(**kwargs)

    def register_cls(self, **kwargs):
        self.http.register_cls(**kwargs)

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if (not self.long_running) != (not old.long_running):
                diff["long_running"] = "New long_running" if self.long_running else "Miss long_running"
            elif self.long_running:
                if lr_diff := self.long_running.diff(old.long_running, level):
                    diff["long_running"] = lr_diff

        if level >= CMDDiffLevelEnum.Structure:
            if self.operation_id != old.operation_id:
                diff["operation_id"] = f"{old.operation_id} != {self.operation_id}"

        if level >= CMDDiffLevelEnum.All:
            if self.description != old.description:
                diff["description"] = f"{old.description} != {self.description}"

        http_diff = self.http.diff(old.http, level)
        if http_diff:
            diff["http"] = http_diff
        return diff


class CMDInstanceCreateOperation(CMDOperation):
    POLYMORPHIC_KEY = "instanceCreate"

    instance_create = PolyModelType(
        CMDInstanceCreateAction,
        allow_subclasses=True,
        required=True,
        serialized_name="instanceCreate",
        deserialize_from="instanceCreate"
    )

    def generate_args(self, ref_args, has_subresource, var_prefix=None):
        return self.instance_create.generate_args(ref_args=ref_args, var_prefix=var_prefix)

    def reformat(self, **kwargs):
        self.instance_create.reformat(**kwargs)

    def register_cls(self, **kwargs):
        self.instance_create.register_cls(**kwargs)


class CMDInstanceUpdateOperation(CMDOperation):
    POLYMORPHIC_KEY = "instanceUpdate"

    # properties as nodes
    instance_update = PolyModelType(
        CMDInstanceUpdateAction,
        allow_subclasses=True,
        required=True,
        serialized_name="instanceUpdate",
        deserialize_from="instanceUpdate"
    )

    def generate_args(self, ref_args, has_subresource, var_prefix=None):
        return self.instance_update.generate_args(ref_args=ref_args, var_prefix=var_prefix)

    def reformat(self, **kwargs):
        self.instance_update.reformat(**kwargs)

    def register_cls(self, **kwargs):
        self.instance_update.register_cls(**kwargs)


class CMDInstanceDeleteOperation(CMDOperation):
    POLYMORPHIC_KEY = "instanceDelete"

    instance_delete = PolyModelType(
        CMDInstanceDeleteAction,
        allow_subclasses=True,
        required=True,
        serialized_name="instanceDelete",
        deserialize_from="instanceDelete"
    )

    def generate_args(self, ref_args, has_subresource, var_prefix=None):
        return self.instance_delete.generate_args(ref_args=ref_args, var_prefix=var_prefix)

    def reformat(self, **kwargs):
        self.instance_delete.reformat(**kwargs)

    def register_cls(self, **kwargs):
        self.instance_delete.register_cls(**kwargs)



from utils import exceptions
import logging

from cli.model.atomic import CLIAtomicClient
from command.model.configuration import CMDClientEndpointsByHttpOperation, CMDClientEndpointsByTemplate, CMDJsonSubresourceSelector
from .az_command_ctx import AzCommandCtx
from .az_command_generator import AzCommandGenerator
from .az_arg_group_generator import AzArgGroupGenerator
from .az_selector_generator import AzJsonSelectorGenerator
from .az_operation_generator import AzHttpOperationGenerator
from utils.plane import PlaneEnum
from utils.case import to_camel_case

logger = logging.getLogger('backend')


class AzClientsGenerator:

    def __init__(self, clients: [CLIAtomicClient]):
        # only clients with cfg will be generated.
        self._clients = sorted([client for client in clients if client.cfg], key=lambda c: c.registered_name)

    def iter_clients(self):
        for client in self._clients:
            yield AzClientGenerator(client)

    def is_empty(self):
        return len(self._clients) == 0


class AzClientGenerator:

    def __init__(self, client):
        self._client = client
        self.cmd_ctx = AzCommandCtx()
        if self._client.cfg.arg_group:
            # Register client arguments in _cmd_ctx
            AzArgGroupGenerator(AzCommandGenerator.ARGS_SCHEMA_NAME, self.cmd_ctx, self._client.cfg.arg_group)
        if isinstance(self._client.cfg.endpoints, CMDClientEndpointsByTemplate):
            self.endpoints = AzClientEndpointsByTemplateGenerator(self._client.cfg.endpoints)
        elif isinstance(self._client.cfg.endpoints, CMDClientEndpointsByHttpOperation):
            self.endpoints = AzClientEndpointsByHttpOperationGenerator(self._client.cfg.endpoints, self.cmd_ctx)
        else:
            raise NotImplementedError()

    @property
    def registered_name(self):
        return self._client.registered_name

    @property
    def cls_name(self):
        return self._client.cls_name

    @property
    def aad_scopes(self):
        return self._client.cfg.auth.aad.scopes

    @property
    def helper_cls_name(self):
        return f'_{self.cls_name}Helper'

    def get_arg_clses(self):
        return sorted(self.cmd_ctx.arg_clses.values(), key=lambda a: a.name)

    def get_update_clses(self):
        return sorted(self.cmd_ctx.update_clses.values(), key=lambda s: s.name)

    def get_response_clses(self):
        return sorted(self.cmd_ctx.response_clses.values(), key=lambda s: s.name)


class AzClientEndpointsByTemplateGenerator:

    def __init__(self, endpoints):
        assert isinstance(endpoints, CMDClientEndpointsByTemplate)
        self._endpoints = endpoints

    @property
    def type(self):
        return self._endpoints.type

    @property
    def cloud_metadata(self):
        return self._endpoints.cloud_metadata

    def iter_hosts(self):
        for template in self._endpoints.templates:
            yield template.cloud, template.template


class AzClientEndpointsByHttpOperationGenerator:

    def __init__(self, endpoints, cmd_ctx):
        assert isinstance(endpoints, CMDClientEndpointsByHttpOperation)
        self._endpoints = endpoints
        self._cmd_ctx = cmd_ctx
        if isinstance(endpoints.selector, CMDJsonSubresourceSelector):
            self.selector = AzJsonSelectorGenerator(self._cmd_ctx, endpoints.selector)
        else:
            raise NotImplementedError()
        op_cls_name = to_camel_case(endpoints.operation.operation_id)
        self.operation = AzHttpOperationGenerator(
            op_cls_name,
            self._cmd_ctx,
            endpoints.operation,
            client_endpoints=None,
        )
        # Only arm endpoints is allowed
        assert endpoints.resource.plane == PlaneEnum.Mgmt
        self.client_name = PlaneEnum.http_client(endpoints.resource.plane)

    @property
    def type(self):
        return self._endpoints.type



from schematics.models import Model
from schematics.types import ModelType, ListType
from utils.fields import PlaneField

from ._command_group import CMDCommandGroup
from ._resource import CMDResource


class CMDConfiguration(Model):
    # properties as nodes
    plane = PlaneField(required=True)

    resources = ListType(ModelType(CMDResource), min_size=1, required=True)  # resources contained in configuration file
    command_groups = ListType(
        ModelType(CMDCommandGroup),
        serialized_name='commandGroups',
        deserialize_from='commandGroups',
    )

    class Options:
        serialize_when_none = False

    def reformat(self, **kwargs):
        self.resources = sorted(self.resources, key=lambda r: r.id)
        for group in self.command_groups:
            group.reformat(**kwargs)
        self.command_groups = sorted(
            [group for group in self.command_groups if group.commands or group.command_groups],
            key=lambda g: g.name
        )

    def link(self):
        """Link cls reference with definition"""
        for group in self.command_groups:
            group.link()



class CMDDiffLevelEnum:

    BreakingChange = 1  # diff breaking change part
    Structure = 2    #
    Associate = 5    # include diff for links
    All = 10    # including description and help messages


class CMDArgBuildPrefix:
    Query = '$Query'
    Header = '$Header'
    Path = '$Path'
    ClientEndpoint = '$Client.Endpoint'  # It's used for client config endpoints related args


DEFAULT_CONFIRMATION_PROMPT = "Are you sure you want to perform this operation?"


class CMDBuildInVariants:

    Instance = "$Instance"
    # It's used in client config which endpoints is by http operation
    EndpointInstance = "$EndpointInstance"

    Subresource = "$Subresource"
    Endpoint = "$Endpoint"



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from cli.model.common._fields import CLICommandNameField, CLIVersionField, CLIModifiedField
from schematics.models import Model
from schematics.types import ListType, BooleanType


class CLIViewCommand(Model):
    names = ListType(field=CLICommandNameField(), min_size=1, required=True)  # full name of a command
    registered = BooleanType()  # register in command table or not

    version = CLIVersionField()  # the version of wait command is not required.
    modified = CLIModifiedField()

    class Options:
        serialize_when_none = False



from .cfg_reader import CfgReader


class CfgValidator:

    def __init__(self, cfg_reader):
        assert isinstance(cfg_reader, CfgReader)
        self.cfg_reader = cfg_reader

    def verify(self):
        # TODO:
        return True

    def _verify_command(self):
        pass

    def _verify_command_arguments(self):
        pass

    def _verify_command_operations(self):
        pass

    def _verify_command_outputs(self):
        pass



import os

from flask import Blueprint, jsonify, request, url_for, redirect

from command.controller.workspace_manager import WorkspaceManager
from utils import exceptions
from utils.config import Config
from command.model.configuration._utils import CMDArgBuildPrefix

bp = Blueprint('editor', __name__, url_prefix='/AAZ/Editor')


@bp.route("/Workspaces", methods=("GET", "POST"))
def editor_workspaces():
    if request.method == "POST":
        # create a new workspace
        # the name of workspace is required
        data = request.get_json()
        if not data or not isinstance(data, dict) or 'name' not in data or 'plane' not in data or 'modNames' not in data or 'resourceProvider' not in data or 'source' not in data:
            raise exceptions.InvalidAPIUsage("Invalid request body")
        name = data['name']
        plane = data['plane']
        mod_names = data['modNames']
        resource_provider = data['resourceProvider']
        source = data['source']
        manager = WorkspaceManager.new(name, plane, mod_names, resource_provider, source)
        manager.save()
        result = manager.ws.to_primitive()
        result.update({
            'url': url_for('editor.editor_workspace', name=manager.name),
            'folder': manager.folder,
            'updated': os.path.getmtime(manager.path)
        })
    elif request.method == "GET":
        result = []
        for ws in WorkspaceManager.list_workspaces():
            result.append({
                **ws,
                'url': url_for('editor.editor_workspace', name=ws['name']),
            })
    else:
        raise NotImplementedError(request.method)

    return jsonify(result)


@bp.route("/Workspaces/<name>", methods=("GET", "DELETE"))
def editor_workspace(name):
    manager = WorkspaceManager(name)
    if request.method == "GET":
        manager.load()
    elif request.method == "DELETE":
        if manager.delete():
            return '', 200
        else:
            return '', 204  # resource not found
    else:
        raise NotImplementedError()

    result = manager.ws.to_primitive()
    result.update({
        'url': url_for('editor.editor_workspace', name=manager.name),
        'folder': manager.folder,
        'updated': os.path.getmtime(manager.path)
    })
    return jsonify(result)


@bp.route("/Workspaces/<name>/SwaggerDefault", methods=("GET",))
def get_workspace_swagger_default_options(name):
    manager = WorkspaceManager(name)
    manager.load()
    result = {
        "plane": manager.ws.plane,
        "source": manager.ws.source,
        "modNames": manager.ws.mod_names if manager.ws.mod_names else Config.DEFAULT_SWAGGER_MODULE,
        "rpName": manager.ws.resource_provider if manager.ws.resource_provider else Config.DEFAULT_RESOURCE_PROVIDER,
    }
    if result["modNames"]:
        result["modNames"] = result["modNames"].split('/')
    return jsonify(result)


@bp.route("/Workspaces/<name>/Rename", methods=("POST",))
def rename_workspace(name):
    manager = WorkspaceManager(name)
    if request.method == "POST":
        data = request.get_json()
        if 'name' not in data or not data['name']:
            raise exceptions.InvalidAPIUsage("Invalid request")
        new_name = data['name'].strip()
        manager.rename(new_name)
        result = manager.ws.to_primitive()
        result.update({
            'url': url_for('editor.editor_workspace', name=manager.name),
            'folder': manager.folder,
            'updated': os.path.getmtime(manager.path)
        })
    else:
        raise NotImplementedError()
    return jsonify(result)


@bp.route("/Workspaces/<name>/Generate", methods=("POST",))
def editor_workspace_generate(name):
    manager = WorkspaceManager(name)
    manager.load()
    manager.generate_to_aaz()
    return "", 200


# client configuration
@bp.route("/Workspaces/<name>/ClientConfig", methods=("GET", "POST"))
def editor_workspace_client_config(name):
    manager = WorkspaceManager(name)
    manager.load()
    if request.method == "GET":
        cfg_editor = manager.load_client_cfg_editor()
        if not cfg_editor:
            raise exceptions.ResourceNotFind("Client configuration not exist")
    elif request.method == "POST":
        data = request.get_json()
        if 'auth' not in data:
            raise exceptions.InvalidAPIUsage("Invalid request: auth info is required.")
        if 'templates' not in data and 'resource' not in data:
            raise exceptions.InvalidAPIUsage("Invalid request: templates or resource is required for endpoints.")
        if 'resource' in data:
            if 'id' not in data['resource'] or 'version' not in data['resource'] or 'module' not in data['resource'] or 'subresource' not in data['resource']:
                raise exceptions.InvalidAPIUsage("Invalid request")
        cfg_editor = manager.create_cfg_editor(
            auth=data['auth'],
            templates=data.get('templates', None),
            cloud_medadata=data.get('cloudMetadata', None),
            arm_resource=data.get('resource', None),
        )
        manager.save()
    else:
        raise NotImplementedError()
    result = cfg_editor.cfg.to_primitive()
    return jsonify(result)


@bp.route("/Workspaces/<name>/ClientConfig/AAZ/Compare", methods=("POST",))
def compare_workspace_client_config_version_with_aaz(name):
    manager = WorkspaceManager(name)
    manager.load()
    if not manager.compare_client_cfg_with_spec():
        raise exceptions.ResourceConflict("Client configuration is out of data in current workspace. Please reload it from aaz repo.")
    return "", 200


@bp.route("/Workspaces/<name>/ClientConfig/AAZ/Inherit", methods=("POST",))
def inherit_workspace_client_config_from_aaz(name):
    manager = WorkspaceManager(name)
    manager.load()
    manager.inherit_client_cfg_from_spec()
    manager.save()
    cfg_editor = manager.load_client_cfg_editor()
    if not cfg_editor:
        raise exceptions.ResourceNotFind("Client configuration not exist")
    result = cfg_editor.cfg.to_primitive()
    return jsonify(result)


@bp.route("/Workspaces/<name>/ClientConfig/Arguments/<arg_var>", methods=("GET", "PATCH"))
def editor_workspace_client_config_argument(name, arg_var):
    manager = WorkspaceManager(name)
    manager.load()
    cfg_editor = manager.load_client_cfg_editor()
    if not cfg_editor:
        raise exceptions.ResourceNotFind("Client configuration not exist")
    arg = cfg_editor.find_arg_by_var(arg_var=arg_var)
    if not arg:
        raise exceptions.ResourceNotFind("Argument not exist")

    if request.method == "GET":
        result = arg.to_primitive()
    elif request.method == "PATCH":
        data = request.get_json()
        cfg_editor.update_arg_by_var(arg_var=arg_var, **data)
        manager.save()
        arg = cfg_editor.find_arg_by_var(arg_var=arg_var)
        result = arg.to_primitive()
    else:
        raise NotImplementedError()
    return jsonify(result)


# command tree operations
@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>", methods=("GET", "POST", "PATCH", "DELETE"))
def editor_workspace_command_tree_node(name, node_names):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command group not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    node = manager.find_command_tree_node(*node_names)
    if not node and request.method != "DELETE":
        raise exceptions.ResourceNotFind("Command group not exist")

    if request.method == "GET":
        # get current node
        result = node.to_primitive()
    elif request.method == "POST":
        # create sub node
        data = request.get_json()
        if 'name' not in data or not data['name']:
            raise exceptions.InvalidAPIUsage("Invalid request")
        sub_node_names = data['name'].split(' ')
        node = manager.create_command_tree_nodes(*node_names, *sub_node_names)
        manager.save()
        result = node.to_primitive()
    elif request.method == "PATCH":
        # update help or stage of node
        data = request.get_json()
        if 'help' in data:
            node = manager.update_command_tree_node_help(*node_names, help=data['help'])
        if 'stage' in data and node.stage != data['stage']:
            node = manager.update_command_tree_node_stage(*node_names, stage=data['stage'])
        manager.save()
        result = node.to_primitive()
    elif request.method == "DELETE":
        # delete node
        if len(node_names) < 1:
            raise exceptions.InvalidAPIUsage("Not support to delete command tree root")
        if not manager.delete_command_tree_node(*node_names):
            return '', 204  # resource not found
        manager.save()
        return '', 200
    else:
        raise NotImplementedError()
    return jsonify(result)


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Rename", methods=("POST",))
def editor_workspace_command_tree_node_rename(name, node_names):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command group not exist")
    node_names = node_names[1:]
    if not node_names:
        raise exceptions.InvalidAPIUsage("Cannot Rename root node")

    manager = WorkspaceManager(name)
    manager.load()
    if not manager.find_command_tree_node(*node_names):
        raise exceptions.ResourceNotFind("Command group not exist")

    data = request.get_json()
    new_name = data.get("name", None)
    if not new_name or not isinstance(new_name, str):
        raise exceptions.InvalidAPIUsage("Invalid request")

    new_node_names = new_name.split(' ')
    node = manager.rename_command_tree_node(*node_names, new_node_names=new_node_names)
    result = node.to_primitive()
    manager.save()
    return jsonify(result)


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/GenerateExamples",
          methods=["POST"])
def editor_workspace_generate_examples(name, node_names, leaf_name):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist.")

    manager = WorkspaceManager(name)
    manager.load()

    node_names = node_names[1:]
    leaf = manager.find_command_tree_leaf(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist.")

    data = request.get_json()
    source = data.get("source", None)

    cfg_editor = manager.load_cfg_editor_by_command(leaf)
    command = cfg_editor.find_command(*leaf.names)
    if not command:
        raise exceptions.ResourceNotFind("Command not exist.")

    if source == "swagger":
        examples = manager.generate_examples_by_swagger(leaf, command)
        result = [example.to_primitive() for example in examples]
    else:
        raise exceptions.InvalidAPIUsage("Invalid request.")

    return jsonify(result)


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Examples",
          methods=["PATCH"])
def editor_workspace_examples(name, node_names, leaf_name):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    leaf = manager.find_command_tree_leaf(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist")

    data = request.get_json()
    if 'examples' in data:
        leaf = manager.update_command_tree_leaf_examples(*leaf.names, examples=data['examples'])
    cfg_editor = manager.load_cfg_editor_by_command(leaf)

    command = cfg_editor.find_command(*leaf.names)
    result = command.to_primitive()
    manager.save()

    del result['name']
    result.update({
        'names': leaf.names,
        'help': leaf.help.to_primitive(),
        'stage': leaf.stage,
    })
    if leaf.examples:
        result['examples'] = [e.to_primitive() for e in leaf.examples]

    return jsonify(result)


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>",
          methods=("GET", "PATCH"))
def editor_workspace_command(name, node_names, leaf_name):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    leaf = manager.find_command_tree_leaf(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist")

    if request.method == "GET":
        # get the command configuration
        cfg_editor = manager.load_cfg_editor_by_command(leaf)
        command = cfg_editor.find_command(*leaf.names)
        result = command.to_primitive()
    elif request.method == "PATCH":
        # update help or stage of node
        data = request.get_json()
        if 'help' in data:
            leaf = manager.update_command_tree_leaf_help(*leaf.names, help=data['help'])
        if 'stage' in data and leaf.stage != data['stage']:
            leaf = manager.update_command_tree_leaf_stage(*leaf.names, stage=data['stage'])
        if 'examples' in data:
            leaf = manager.update_command_tree_leaf_examples(*leaf.names, examples=data['examples'])
        cfg_editor = manager.load_cfg_editor_by_command(leaf)
        if 'confirmation' in data:
            cfg_editor.update_command_confirmation(*leaf.names, confirmation=data['confirmation'])
        if 'outputs' in data:
            cfg_editor.update_command_outputs(*leaf.names, outputs=data['outputs'])
        command = cfg_editor.find_command(*leaf.names)
        result = command.to_primitive()
        manager.save()
    else:
        raise NotImplementedError()

    del result['name']
    result.update({
        'names': leaf.names,
        'help': leaf.help.to_primitive(),
        'stage': leaf.stage,
    })
    if leaf.examples:
        result['examples'] = [e.to_primitive() for e in leaf.examples]

    # add client configuration argument group in the command response
    client_cfg = manager.load_client_cfg_editor()
    if client_cfg and client_cfg.cfg.arg_group:
        arg_groups = result.get('argGroups', [])
        result['argGroups'] = [*arg_groups, client_cfg.cfg.arg_group.to_primitive()]

    return jsonify(result)


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Rename",
          methods=("POST",))
def editor_workspace_command_rename(name, node_names, leaf_name):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    if not manager.find_command_tree_leaf(*node_names, leaf_name):
        raise exceptions.ResourceNotFind("Command not exist")

    data = request.get_json()
    new_name = data.get("name", None)
    if not new_name or not isinstance(new_name, str):
        raise exceptions.InvalidAPIUsage("Invalid request")

    new_leaf_names = new_name.split(' ')
    new_leaf = manager.rename_command_tree_leaf(*node_names, leaf_name, new_leaf_names=new_leaf_names)
    cfg_editor = manager.load_cfg_editor_by_command(new_leaf)
    command = cfg_editor.find_command(*new_leaf.names)

    result = command.to_primitive()
    del result['name']
    result.update({
        'names': new_leaf.names,
        'help': new_leaf.help.to_primitive(),
        'stage': new_leaf.stage
    })
    if new_leaf.examples:
        result['examples'] = [e.to_primitive() for e in new_leaf.examples]

    manager.save()
    return jsonify(result)


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Arguments/<arg_var>",
          methods=("GET", "PATCH"))
def editor_workspace_command_argument(name, node_names, leaf_name, arg_var):
    if arg_var.startswith(CMDArgBuildPrefix.ClientEndpoint):
        # redirect to client config argument
        return redirect(url_for('editor.editor_workspace_client_config_argument', name=name, arg_var=arg_var))

    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    leaf = manager.find_command_tree_leaf(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist")
    cfg_editor = manager.load_cfg_editor_by_command(leaf)
    arg, _ = cfg_editor.find_arg_by_var(*node_names, leaf_name, arg_var=arg_var)
    if not arg:
        raise exceptions.ResourceNotFind("Argument not exist")

    if request.method == "GET":
        result = arg.to_primitive()
    elif request.method == "PATCH":
        data = request.get_json()
        cfg_editor.update_arg_by_var(*node_names, leaf_name, arg_var=arg_var, **data)
        manager.save()
        arg, _ = cfg_editor.find_arg_by_var(*node_names, leaf_name, arg_var=arg_var)
        result = arg.to_primitive()
    else:
        raise NotImplementedError()
    return jsonify(result)


# TODO: support to modify element of array or dict arguments
# @bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Arguments/<arg_var>/Element",
#           methods=("GET", "PATCH"))
# def editor_workspace_command_arg_element_of_array_or_dict(name, node_names, leaf_name, arg_var):
#     pass


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Arguments/<arg_var>/Flatten",
          methods=("POST", ))
def editor_workspace_command_argument_flatten(name, node_names, leaf_name, arg_var):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    leaf = manager.find_command_tree_leaf(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist")
    cfg_editor = manager.load_cfg_editor_by_command(leaf)

    # flatten argument variant
    arg, _ = cfg_editor.find_arg_by_var(*node_names, leaf_name, arg_var=arg_var)
    if not arg:
        raise exceptions.ResourceNotFind("Argument not exit")
    data = request.get_json()
    sub_args_options = data.get('subArgsOptions', None)
    cfg_editor.flatten_arg(*node_names, leaf_name, arg_var=arg_var, sub_args_options=sub_args_options)
    manager.save()

    return '', 200


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Arguments/<arg_var>/Unflatten",
          methods=("POST", ))
def editor_workspace_command_argument_unflatten(name, node_names, leaf_name, arg_var):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    leaf = manager.find_command_tree_leaf(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist")
    cfg_editor = manager.load_cfg_editor_by_command(leaf)

    parent, arg, _ = cfg_editor.find_arg_with_parent_by_var(*node_names, leaf_name, arg_var=arg_var)
    if arg:
        raise exceptions.ResourceConflict("Argument already exit")
    elif not parent:
        raise exceptions.ResourceNotFind("Argument not able to flatten")

    data = request.get_json()
    sub_args_options = data.get('subArgsOptions', None)
    cfg_editor.unflatten_arg(*node_names, leaf_name, arg_var=arg_var, options=data['options'], help=data['help'],
                             sub_args_options=sub_args_options)
    manager.save()
    arg, _ = cfg_editor.find_arg_by_var(*node_names, leaf_name, arg_var=arg_var)
    result = arg.to_primitive()

    return jsonify(result)


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Arguments/<arg_var>/UnwrapClass",
          methods=("POST", ))
def editor_workspace_command_argument_unwrap_class(name, node_names, leaf_name, arg_var):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    leaf = manager.find_command_tree_leaf(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist")
    cfg_editor = manager.load_cfg_editor_by_command(leaf)

    # unwrap argument variant
    arg, _ = cfg_editor.find_arg_by_var(*node_names, leaf_name, arg_var=arg_var)
    if not arg:
        raise exceptions.ResourceNotFind("Argument not exit")
    cfg_editor.unwrap_cls_arg(*node_names, leaf_name, arg_var=arg_var)
    manager.save()
    return '', 200


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Arguments/<arg_var>/FindSimilar",
          methods=("POST",))
def editor_workspace_command_argument_find_similar(name, node_names, leaf_name, arg_var):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    leaf = manager.find_command_tree_leaf(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist")
    cfg_editor = manager.load_cfg_editor_by_command(leaf)

    arg, _ = cfg_editor.find_arg_by_var(*node_names, leaf_name, arg_var=arg_var)
    if not arg:
        raise exceptions.ResourceNotFind("Argument not exist")

    result = {
        WorkspaceManager.COMMAND_TREE_ROOT_NAME: {
            "id": url_for('editor.editor_workspace_command_tree_node',
                          name=name,
                          node_names=[WorkspaceManager.COMMAND_TREE_ROOT_NAME])
        }
    }
    for cmd_names, args_map in manager.find_similar_args(*leaf.names, arg=arg).items():
        node = result[WorkspaceManager.COMMAND_TREE_ROOT_NAME]
        for idx, group_name in enumerate(cmd_names[:-1]):
            if 'commandGroups' not in node:
                node['commandGroups'] = {}
            if group_name not in node['commandGroups']:
                node['commandGroups'][group_name] = {
                    "id": url_for('editor.editor_workspace_command_tree_node',
                                  name=name,
                                  node_names=[WorkspaceManager.COMMAND_TREE_ROOT_NAME, *cmd_names[:idx+1]])
                }
            node = node['commandGroups'][group_name]
        if 'commands' not in node:
            node['commands'] = {}
        if cmd_names[-1] not in node['commands']:
            node['commands'][cmd_names[-1]] = {
                "id": url_for('editor.editor_workspace_command',
                              name=name,
                              node_names=[WorkspaceManager.COMMAND_TREE_ROOT_NAME, *cmd_names[:-1]],
                              leaf_name=cmd_names[-1]),
                "names": cmd_names,
                "args": {}
            }
        args = node['commands'][cmd_names[-1]]['args']
        for arg_var, arg_idx_list in args_map.items():
            if arg_var not in args:
                args[arg_var] = []
            args[arg_var].extend(arg_idx_list)
            args[arg_var] = sorted(set(args[arg_var]))
    return jsonify(result)


# command tree resource operations
@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/AddSwagger", methods=("POST",))
def editor_workspace_tree_node_add_swagger_resources(name, node_names):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command group not exist")
    node_names = node_names[1:]
    if len(node_names) > 0:
        raise exceptions.InvalidAPIUsage("Not support to add resources under a specific node.")

    manager = WorkspaceManager(name)
    manager.load()
    if not manager.find_command_tree_node(*node_names):
        raise exceptions.ResourceNotFind("Command group not exist")

    # add new resource
    data = request.get_json()
    if not isinstance(data, dict):
        raise exceptions.InvalidAPIUsage("Invalid request")

    try:
        mod_names = data['module']
        version = data['version']
        resources = data['resources']
    except KeyError:
        raise exceptions.InvalidAPIUsage("Invalid request")

    manager.add_new_resources_by_swagger(
        mod_names=mod_names,
        version=version,
        resources=resources,
    )
    manager.save()
    return "", 200


# command tree resource operations
@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/AddTypespec", methods=("POST",))
def editor_workspace_tree_node_add_typespec_resources(name, node_names):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command group not exist")
    node_names = node_names[1:]
    if len(node_names) > 0:
        raise exceptions.InvalidAPIUsage("Not support to add resources under a specific node.")

    manager = WorkspaceManager(name)
    manager.load()
    if not manager.find_command_tree_node(*node_names):
        raise exceptions.ResourceNotFind("Command group not exist")

    # add new resource
    data = request.get_json()
    if not isinstance(data, dict):
        raise exceptions.InvalidAPIUsage("Invalid request")

    try:
        version = data['version']
        resources = data['resources']
    except KeyError:
        raise exceptions.InvalidAPIUsage("Invalid request")

    manager.add_new_resources_by_typespec(
        version=version,
        resources=resources,
    )
    manager.save()
    return "", 200


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Resources", methods=("GET",))
def editor_workspace_resources(name, node_names):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command group not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    if not manager.find_command_tree_node(*node_names):
        raise exceptions.ResourceNotFind("Command group not exist")

    resources = manager.get_resources(*node_names)

    result = [r.to_primitive() for r in resources]
    return jsonify(result)


@bp.route("/Workspaces/<name>/Resources/Merge", methods=("POST",))
def editor_workspace_resources_merge(name):
    manager = WorkspaceManager(name)
    manager.load()
    data = request.get_json()
    if "mainResource" not in data or "plusResource" not in data:
        raise exceptions.InvalidAPIUsage("Invalid request")
    main_resource_id = data["mainResource"]["id"]
    main_resource_version = data["mainResource"]["version"]
    plus_resource_id = data["plusResource"]["id"]
    plus_resource_version = data["plusResource"]["version"]
    if not manager.merge_resources(main_resource_id, main_resource_version, plus_resource_id, plus_resource_version):
        raise exceptions.ResourceConflict("Cannot merge resources")
    manager.save()
    return "", 200


@bp.route("/Workspaces/<name>/Resources/ReloadSwagger", methods=("POST",))
def editor_workspace_resource_reload_swagger(name):
    # update resource by reloading swagger
    manager = WorkspaceManager(name)
    manager.load()
    data = request.get_json()
    try:
        resources = data['resources']
    except KeyError:
        raise exceptions.InvalidAPIUsage("Invalid request")
    manager.reload_resources_by_swagger(resources=resources)
    manager.save()
    return "", 200


@bp.route("/Workspaces/<name>/Resources/ReloadTypespec", methods=("POST",))
def editor_workspace_resource_reload_typespec(name):
    # update resource by reloading typespec
    manager = WorkspaceManager(name)
    manager.load()
    data = request.get_json()
    try:
        resources = data['resources']
    except KeyError:
        raise exceptions.InvalidAPIUsage("Invalid request")
    manager.reload_resources_by_typespec(resources=resources)
    manager.save()
    return "", 200


@bp.route("/Workspaces/<name>/Resources/<base64:resource_id>/V/<base64:version>", methods=("DELETE",))
def editor_workspace_resource(name, resource_id, version):
    # remove commands of the resource, including commands of subresources
    manager = WorkspaceManager(name)
    manager.load()
    if not manager.remove_resource(resource_id, version):
        return "", 204
    manager.save()
    return "", 200


@bp.route("/Workspaces/<name>/Resources/<base64:resource_id>/V/<base64:version>/Commands", methods=("GET",))
def list_workspace_resource_related_commands(name, resource_id, version):
    # list commands of the resource, including commands of sub resources
    manager = WorkspaceManager(name)
    manager.load()
    commands = manager.list_commands_by_resource(resource_id, version)
    result = [command.to_primitive() for command in commands]
    return jsonify(result)


@bp.route("/Workspaces/<name>/Resources/<base64:resource_id>/V/<base64:version>/Subresources", methods=("POST",))
def editor_workspace_subresources(name, resource_id, version):
    # add subresource command
    manager = WorkspaceManager(name)
    manager.load()
    data = request.get_json()
    try:
        arg_var = data['arg']
        cg_names = [nm for nm in data['commandGroupName'].split(' ') if nm]
        ref_args_options = data.get('refArgsOptions', None)
    except KeyError:
        raise exceptions.InvalidAPIUsage("Invalid request")
    manager.add_subresource_by_arg_var(resource_id, version, arg_var, cg_names, ref_args_options)
    manager.save()
    return "", 200


@bp.route("/Workspaces/<name>/Resources/<base64:resource_id>/V/<base64:version>/Subresources/<base64:subresource>", methods=("DELETE",))
def editor_workspace_subresource(name, resource_id, version, subresource):
    # Remove commands of subresource
    manager = WorkspaceManager(name)
    manager.load()
    if not manager.remove_subresource(resource_id, version, subresource):
        return "", 204
    manager.save()
    return "", 200


@bp.route("/Workspaces/<name>/Resources/<base64:resource_id>/V/<base64:version>/Subresources/<base64:subresource>/Commands", methods=("GET",))
def list_workspace_subresource_related_commands(name, resource_id, version, subresource):
    # list commands of subresource
    manager = WorkspaceManager(name)
    manager.load()
    commands = manager.list_commands_by_subresource(resource_id, version, subresource)
    result = [command.to_primitive() for command in commands]
    return jsonify(result)


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Try", methods=("POST",))
def editor_workspace_try_command_group(name, node_names):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command group not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    if not manager.find_command_tree_node(*node_names):
        raise exceptions.ResourceNotFind("Command group not exist")

    # try sub commands by installed as a try extension of cli
    raise NotImplementedError()


@bp.route("/Workspaces/<name>/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Try", methods=("POST",))
def editor_workspace_try_command(name, node_names, leaf_name):
    if node_names[0] != WorkspaceManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = WorkspaceManager(name)
    manager.load()
    if not manager.find_command_tree_leaf(*node_names, leaf_name):
        raise exceptions.ResourceNotFind("Command not exist")

    # try command by installed as a try extension of cli
    raise NotImplementedError()



from schematics.models import Model
from schematics.types import StringType, FloatType, IntType
from ._fields import CMDRegularExpressionField, CMDBooleanField
from ._utils import CMDDiffLevelEnum


class CMDFormat(Model):

    class Options:
        serialize_when_none = False

    def build_arg_fmt(self, builder, ref_fmt):
        """Build argument format from schema format"""
        raise NotImplementedError()


# string
class CMDStringFormat(CMDFormat):
    pattern = CMDRegularExpressionField()
    max_length = IntType(
        serialized_name="maxLength",
        deserialize_from="maxLength",
        min_value=0
    )
    min_length = IntType(
        serialized_name="minLength",
        deserialize_from="minLength",
        min_value=0
    )

    def build_arg_fmt(self, builder, ref_fmt):
        fmt = CMDStringFormat()
        fmt.pattern = self.pattern
        fmt.max_length = self.max_length
        fmt.min_length = self.min_length
        return fmt

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            # skip pattern difference, because it's hard to tell
            if self.max_length and (not old.max_length or old.max_length > self.max_length):
                diff["max_length"] = f"from {old.max_length} to {self.max_length}"
            if self.min_length and (not old.min_length or old.min_length < self.min_length):
                diff["min_length"] = f"from {old.min_length} to {self.min_length}"

        if level >= CMDDiffLevelEnum.Structure:
            if self.pattern != old.pattern:
                diff["pattern"] = f"{old.pattern} != {self.pattern}"
            if self.max_length != old.max_length:
                diff["max_length"] = f"{old.max_length} != {self.max_length}"
            if self.min_length != old.min_length:
                diff["min_length"] = f"{old.min_length} != {self.min_length}"
        return diff


class CMDResourceIdFormat(CMDFormat):
    template = StringType(required=True)

    def build_arg_fmt(self, builder, ref_fmt):
        fmt = CMDResourceIdFormat()
        fmt.template = self.template
        return fmt

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.template != old.template:
                diff['template'] = f"from {old.template} to {self.template}"
        return diff


# integer
class CMDIntegerFormat(CMDFormat):
    multiple_of = IntType(
        min_value=0,
        serialized_name='multipleOf',
        deserialize_from='multipleOf'
    )
    maximum = IntType()
    minimum = IntType()

    def build_arg_fmt(self, builder, ref_fmt):
        fmt = CMDIntegerFormat()
        fmt.multiple_of = self.multiple_of
        fmt.maximum = self.maximum
        fmt.minimum = self.minimum
        return fmt

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}

        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.multiple_of and (not old.multiple_of or
                                     self.multiple_of != 0 and old.multiple_of % self.multiple_of != 0):
                diff["multiple_of"] = f"from {old.multiple_of} to {self.multiple_of}"
            if self.maximum and (not old.maximum or old.maximum > self.maximum):
                diff["maximum"] = f"from {old.maximum} to {self.maximum}"
            if self.minimum and (not old.minimum or old.minimum < self.minimum):
                diff["minimum"] = f"from {old.minimum} to {self.minimum}"

        if level >= CMDDiffLevelEnum.Structure:
            if self.multiple_of != old.multiple_of:
                diff["multiple_of"] = f"{old.multiple_of} != {self.multiple_of}"
            if self.maximum != old.maximum:
                diff["maximum"] = f"{old.maximum} != {self.maximum}"
            if self.minimum != old.minimum:
                diff["minimum"] = f"{old.minimum} != {self.minimum}"
        return diff


# float
class CMDFloatFormat(CMDFormat):
    multiple_of = FloatType(
        min_value=0,
        serialized_name='multipleOf',
        deserialize_from='multipleOf'
    )
    maximum = FloatType()
    exclusive_maximum = CMDBooleanField(
        serialized_name='exclusiveMaximum',
        deserialize_from='exclusiveMaximum'
    )
    minimum = FloatType()
    exclusive_minimum = CMDBooleanField(
        serialized_name='exclusiveMinimum',
        deserialize_from='exclusiveMinimum'
    )

    def build_arg_fmt(self, builder, ref_fmt):
        fmt = CMDFloatFormat()
        fmt.multiple_of = self.multiple_of
        fmt.maximum = self.maximum
        fmt.exclusive_maximum = self.exclusive_maximum
        fmt.minimum = self.minimum
        fmt.exclusive_minimum = self.exclusive_minimum
        return fmt

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}

        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.multiple_of and (not old.multiple_of or
                                     self.multiple_of != 0 and old.multiple_of % self.multiple_of != 0):
                diff["multiple_of"] = f"from {old.multiple_of} to {self.multiple_of}"
            if self.maximum and (not old.maximum or old.maximum > self.maximum):
                diff["maximum"] = f"from {old.maximum} to {self.maximum}"
            if self.maximum and self.maximum == old.maximum and self.exclusive_maximum != old.exclusive_maximum:
                diff["exclusive_maximum"] = f"from {old.exclusive_maximum} to {self.exclusive_maximum}"
            if self.minimum and (not old.minimum or old.minimum < self.minimum):
                diff["minimum"] = f"from {old.minimum} to {self.minimum}"
            if self.minimum and self.minimum == old.minimum and self.exclusive_minimum != old.exclusive_minimum:
                diff["exclusive_minimum"] = f"from {old.exclusive_minimum} to {self.exclusive_minimum}"

        if level >= CMDDiffLevelEnum.Structure:
            if self.multiple_of != old.multiple_of:
                diff["multiple_of"] = f"{old.multiple_of} != {self.multiple_of}"
            if self.maximum != old.maximum:
                diff["maximum"] = f"{old.maximum} != {self.maximum}"
            if self.exclusive_maximum != old.exclusive_maximum:
                diff["exclusive_maximum"] = f"{old.exclusive_maximum} != {self.exclusive_maximum}"
            if self.minimum != old.minimum:
                diff["minimum"] = f"{old.minimum} != {self.minimum}"
            if self.exclusive_minimum != old.exclusive_minimum:
                diff["exclusive_minimum"] = f"{old.exclusive_minimum} != {self.exclusive_minimum}"
        return diff


# object
class CMDObjectFormat(CMDFormat):
    max_properties = IntType(
        min_value=0,
        serialized_name='maxProperties',
        deserialize_from='maxProperties'
    )
    min_properties = IntType(
        min_value=0,
        serialized_name='minProperties',
        deserialize_from='minProperties'
    )

    def build_arg_fmt(self, builder, ref_fmt):
        fmt = CMDObjectFormat()
        fmt.max_properties = self.max_properties
        fmt.min_properties = self.min_properties
        return fmt

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}

        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.max_properties and (not old.max_properties or old.max_properties > self.max_properties):
                diff["max_properties"] = f"from {old.max_properties} to {self.max_properties}"
            if self.min_properties and (not old.min_properties or old.min_properties < self.min_properties):
                diff["min_properties"] = f"from {old.min_properties} to {self.min_properties}"

        if level >= CMDDiffLevelEnum.Structure:
            if self.max_properties != old.max_properties:
                diff["max_properties"] = f"{old.max_properties} != {self.max_properties}"
            if self.min_properties != old.min_properties:
                diff["min_properties"] = f"{old.min_properties} != {self.min_properties}"
        return diff


# array
class CMDArrayFormat(CMDFormat):
    unique = CMDBooleanField()
    max_length = IntType(
        min_value=0,
        serialized_name='maxLength',
        deserialize_from='maxLength'
    )
    min_length = IntType(
        min_value=0,
        serialized_name='minLength',
        deserialize_from='minLength'
    )

    str_format = StringType(
        choices=(
            "csv",  # default
            "ssv",
            "tsv",
            "pipes",
            "multi"
        ),
        serialized_name="strFormat",
        deserialize_from="strFormat",
    )   # the format convert an array instance to a string

    def build_arg_fmt(self, builder, ref_fmt):
        fmt = CMDArrayFormat()
        fmt.unique = self.unique
        fmt.max_length = self.max_length
        fmt.min_length = self.min_length
        # ignore str_format, it's not required for argument
        return fmt

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}

        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.unique != old.unique:
                diff["unique"] = f"{old.unique} != {self.unique}"
            if self.str_format != old.str_format:
                diff["str_format"] = f"{old.str_format} != {self.str_format}"
            if self.max_length and (not old.max_length or old.max_length > self.max_length):
                diff["max_length"] = f"from {old.max_length} to {self.max_length}"
            if self.min_length and (not old.min_length or old.min_length < self.min_length):
                diff["min_length"] = f"from {old.min_length} to {self.min_length}"

        if level >= CMDDiffLevelEnum.Structure:
            if self.max_length != old.max_length:
                diff["max_length"] = f"{old.max_length} != {self.max_length}"
            if self.min_length != old.min_length:
                diff["min_length"] = f"{old.min_length} != {self.min_length}"
        return diff



import logging

from command.model.configuration import CMDResourceGroupNameArg
from utils import exceptions
from utils.case import to_snake_case
from .az_arg_group_generator import AzArgClsGenerator
from .az_operation_generator import AzRequestClsGenerator, AzResponseClsGenerator

logger = logging.getLogger('backend')


class AzCommandCtx:

    def __init__(self):
        self._cls_arg_maps = {}
        self._ctx_arg_map = {}
        self._selectors = {}
        self.rg_arg_var = None

        self.arg_clses = {}
        self.update_clses = {}
        self.response_clses = {}
        self.support_id_part = True

    def set_argument_cls(self, arg):
        cls_name = arg.cls
        self._cls_arg_maps[f"@{cls_name}"] = {}
        assert cls_name not in self.arg_clses, f"Argument class {cls_name} is defined more than one place"
        self.arg_clses[cls_name] = AzArgClsGenerator(cls_name, self, arg)

    def set_argument(self, keys, arg, ctx_namespace='self.ctx.args'):
        var_name = arg.var
        hide = arg.hide
        if var_name.startswith('@'):
            map_name = var_name.replace('[', '.[').replace('{', '.{').split('.', maxsplit=1)[0]
            if map_name != keys[0]:
                raise exceptions.VerificationError(
                    "Invalid argument var",
                    details=f"argument var '{var_name}' does not start with '{keys[0]}'"
                )
            if map_name not in self._cls_arg_maps:
                self._cls_arg_maps[map_name] = {}
            self._cls_arg_maps[map_name][var_name] = (
                '.'.join(keys).replace('.[', '[').replace('.{', '{'),
                hide
            )
        else:
            self._ctx_arg_map[var_name] = (
                '.'.join([ctx_namespace, *keys]).replace('.[', '[').replace('.{', '{'),
                hide
            )

        if isinstance(arg, CMDResourceGroupNameArg):
            assert self.rg_arg_var is None, "Resource Group Argument defined twice"
            self.rg_arg_var = arg.var

    def get_argument(self, var_name):
        if var_name.startswith('@'):
            map_name = var_name.replace('[', '.[').replace('{', '.{').split('.', maxsplit=1)[0]
            if map_name not in self._cls_arg_maps:
                raise exceptions.VerificationError(
                    "Invalid argument var",
                    details=f"argument var '{var_name}' has unregistered class '{map_name}'."
                )
            if var_name not in self._cls_arg_maps[map_name]:
                raise exceptions.VerificationError(
                    "Invalid argument var",
                    details=f"argument var '{var_name}' does not find."
                )
            return self._cls_arg_maps[map_name][var_name]
        else:
            if var_name not in self._ctx_arg_map:
                raise exceptions.VerificationError(
                    "Invalid argument var",
                    details=f"argument var '{var_name}' does not find."
                )
            return self._ctx_arg_map[var_name]

    def get_variant(self, variant, name_only=False):
        if variant.startswith('$'):
            variant = variant[1:]
        variant = to_snake_case(variant)
        if name_only:
            return variant

        is_selector = variant in self._selectors

        if is_selector:
            return f'self.ctx.selectors.{variant}', is_selector
        else:
            return f'self.ctx.vars.{variant}', is_selector

    def set_update_cls(self, schema):
        cls_name = schema.cls
        assert cls_name not in self.update_clses, f"Schema cls '{cls_name}', is defined more than once"
        self.update_clses[cls_name] = AzRequestClsGenerator(self, cls_name, schema)

    def set_response_cls(self, schema):
        cls_name = schema.cls
        assert cls_name not in self.response_clses, f"Schema cls '{cls_name}', is defined more than once"
        self.response_clses[cls_name] = AzResponseClsGenerator(self, cls_name, schema)

    def set_selector(self, selector):
        self._selectors[self.get_variant(selector.var, name_only=True)] = selector

    def render_arg_resource_id_template(self, template):
        # TODO: fill blank placeholders as much as possible

        return template
